"""Localization module."""
from importlib import import_module

import pluggy
from raesl import logger

pm = pluggy.PluginManager("raesl.doc")
hookspec = pluggy.HookspecMarker("raesl.doc")
hookimpl = pluggy.HookimplMarker("raesl.doc")


class Hookspecs:
    @hookspec(firstresult=True)
    def gettext(key: str):
        """Get text translation."""


pm.add_hookspecs(Hookspecs)


def _(key: str, hook: pluggy.PluginManager = pm.hook):
    """Gettext alike translation function."""
    return hook.gettext(key=key)


def list_locales():
    """List available locales."""
    from pathlib import Path

    here = Path(__file__).parent
    return list(
        p.stem for p in here.glob("*") if p.suffix == ".py" and p.name not in {"__init__.py"}
    )


def register_locale(locale: str = "en"):
    """Register a locale. Existing locales are re-registered."""
    if pm.has_plugin(locale):
        pm.unregister(name=locale)
        logger.debug("Unregistered locale: {}.".format(locale))

    mod = import_module("raesl.doc.locales.{}".format(locale))
    pm.register(mod, name=locale)
    logger.debug("Registered locale: {}.".format(locale))


def register_default_locale(locale: str = "en"):
    """Register default locale if no other locale is set."""
    logger.debug("Registering default locale '{}'...".format(locale))
    if not pm.get_plugins():
        register_locale(locale)
    else:
        logger.debug("Default locale already registered.")



"""ESL Jupyter kernel Command Line Interface."""

import os
import sys
from typing import Optional

import click

kernel_json = {
    "argv": [
        sys.executable,
        "-m",
        "raesl.jupyter.cli",
        "run",
        "-f",
        "{connection_file}",
    ],
    "display_name": "ESL",
    "language": "esl",
    "mimetype": "text/x-esl",
    "file_extension": ".esl",
    "pygments_lexer": "esl",
    "codemirror_mode": "esl",
}


def install_my_kernel_spec(user=True, prefix=None):
    from IPython.utils.tempdir import TemporaryDirectory
    from jupyter_client.kernelspec import KernelSpecManager
    import json

    with TemporaryDirectory() as td:
        print("Installing Jupyter kernel spec...")
        os.chmod(td, 0o755)  # Starts off as 700, not user readable

        with open(os.path.join(td, "kernel.json"), "w") as f:
            json.dump(kernel_json, f, sort_keys=True)

        # assets = pathlib.Path(__file__).parent / "assets"
        # for asset in assets.iterdir():
        #     shutil.copy2(assets / asset, td)

        KernelSpecManager().install_kernel_spec(td, "esl", user=user, prefix=prefix)
        KernelSpecManager().get_kernel_spec("esl")
        print("Done.")


def _is_root():
    try:
        return os.geteuid() == 0
    except AttributeError:
        return False  # assume not an admin on non-Unix platforms


@click.group("jupyter")
def jupyter():
    """Manage the Jupyter ESL kernel."""
    pass


@jupyter.command("install")
@click.option(
    "--user",
    is_flag=True,
    default=False,
    help="Install to the per-user kernels registry. Default if not root.",
)
@click.option(
    "--sys-prefix",
    is_flag=True,
    default=False,
    help="Install to sys.prefix (e.g. a virtualenv or conda env)",
)
@click.option(
    "--prefix",
    default=None,
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    help="Install to the given prefix. "
    "Kernelspec will be installed to {prefix}/share/jupyter/kernels/",
)
def install(user: bool, sys_prefix: bool, prefix: Optional[str] = None) -> None:
    """Install the ESL Jupyter kernel spec."""
    prefix = sys.prefix if sys_prefix else prefix

    if not prefix and not _is_root():
        user = True

    install_my_kernel_spec(user=user, prefix=prefix)


@jupyter.command("run")
@click.option(
    "-f",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    help="Kernel connection file.",
)
def run(f: str):
    """Run the ESL Jupyter kernel."""
    from ipykernel.kernelapp import IPKernelApp
    from raesl.jupyter.kernel import EslKernel

    IPKernelApp.launch_instance(kernel_class=EslKernel)


if __name__ == "__main__":
    jupyter()



"""Class to store and check verb / pre-position definitions."""
import collections
from typing import TYPE_CHECKING, Dict, List, Tuple

from raesl.compile import diagnostics
from raesl.compile.ast import verbs

if TYPE_CHECKING:
    from raesl.compile.ast.specification import Specification
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.ast_builder import AstBuilder


class VerbDefBuilder:
    """Part of the builders to deal with verbs / pre-positions."""

    def __init__(self, ast_builder: "AstBuilder"):
        # Make the builder problem store available locally.
        self.diag_store = ast_builder.diag_store

        # Setup local storage
        self.storage: Dict[Tuple[str, str], List[verbs.VerbPreposDef]]
        self.storage = collections.defaultdict(list)

    def add_verbdef(self, verb_tok: "Token", prepos_tok: "Token"):
        """Store the provided verb/prepos combination."""

        # For error reporting, order storage by the verb and prepos text.
        key = (verb_tok.tok_text.lower(), prepos_tok.tok_text.lower())
        vdef = verbs.VerbPreposDef(verb_tok, prepos_tok)
        self.storage[key].append(vdef)

    def finish(self, spec: "Specification"):
        """Finish collecting by checking the collected verb-prepositions. Store result
        in the provided specification.
        """
        for verb_prepos_text, verbdefs in self.storage.items():
            if len(verbdefs) != 1:
                # More than one definition of the same verb/prepos combination.
                dupes = [vd.verb.get_location() for vd in verbdefs]
                self.diag_store.add(
                    diagnostics.W200(
                        " ".join(verb_prepos_text),
                        "verb-preposition combination",
                        location=dupes[0],
                        dupes=dupes,
                    )
                )

        # spec.verb_prepos = [verbdefs[0] for verbdefs in self.storage.values()]
        # mypy generates false positive on the list comprehension
        spec.verb_prepos = []
        for verbdefs in self.storage.values():
            spec.verb_prepos.append(verbdefs[0])



"""Verb / preposition definitions in ESL."""
from typing import TYPE_CHECKING

from raesl.compile.ast.comment_storage import DefaultDocStore

if TYPE_CHECKING:
    from raesl.compile.scanner import Token


class VerbPreposDef(DefaultDocStore):
    """A verb and a pre-position definition.

    Arguments:
        verb: Verb token.
        prepos: Pre-position token.
    """

    def __init__(self, verb: "Token", prepos: "Token"):
        super(VerbPreposDef, self).__init__(verb)
        self.verb = verb
        self.prepos = prepos

    def __str__(self):
        return "VerbPrepos('{}', '{}')".format(self.verb.tok_text, self.prepos.tok_text)



"""Line matcher state machines for transformations."""
from typing import TYPE_CHECKING

from raesl.compile.ast import components
from raesl.compile.machine_files import sub_clause, typing
from raesl.compile.machine_files.machine_parts import (
    get_argument_references_part,
    get_does_auxiliary_verb_part,
)

if TYPE_CHECKING:
    from raesl.compile.typechecking.ast_builder import AstBuilder


_TRANSFORM_HEADER_SPEC = """
trasform_header:
    start initial;
    start -> s1 [TRANSFORM_REQUIREMENT_KW] tag=transform_kind;
    start -> s1 [TRANSFORM_CONSTRAINT_KW] tag=transform_kind;

    end accept=transform_header;
    s1 -> end [NL_TK];
    s1 -> end [EOF_TK];
"""


def _process_transform_header(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    kind_tok = tags["transform_kind"][0]
    builder.notify_new_section(kind_tok, False)

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.notify_transform_section(kind_tok)
    current_comp.new_transform_header(kind_tok)


_TRANSFORM_MAIN_NO_SUBS_SPEC = (
    """
transform_main_no_subs:
    start initial;
    start -> s1 [NAME] tag=label;
    s1 -> s2 [COLON_TK];
"""
    + get_does_auxiliary_verb_part("s2", ["s4"], tagname="doesaux")
    + """\
    s4 -> s5 [NAME] tag=verb;
"""
    + get_argument_references_part("s5", ["s7"], tagname="from_argument_name", prefix="from")
    + """\
    s7 -> s8 [NAME] tag=prepos;
"""
    + get_argument_references_part("s8", ["s10"], tagname="to_argument_name", prefix="to")
    + """\
    end accept=transform_main_no_subs;
    s10 -> end [NL_TK];
    s10 -> end [EOF_TK];
"""
)


def _process_transform_no_subs(
    tags: typing.TokensDict, _accept: str, builder: "AstBuilder"
) -> None:
    label_tok = tags["label"][0]
    doesaux_tok = tags["doesaux"][0]
    verb_tok = tags["verb"][0]
    in_flows = [components.Flow(name_tok) for name_tok in tags["from_argument_name"]]
    prepos_tok = tags["prepos"][0]
    out_flows = [components.Flow(name_tok) for name_tok in tags["to_argument_name"]]
    transform = components.Transformation(
        label_tok, doesaux_tok, verb_tok, in_flows, prepos_tok, out_flows
    )

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.add_transform(transform)


_TRANSFORM_MAIN_WITH_SUBS_SPEC = (
    """
transform_main_with_subs:
    start initial;
    start -> s1 [NAME] tag=label;
    s1 -> s2 [COLON_TK];
"""
    + get_does_auxiliary_verb_part("s2", ["s4"], tagname="doesaux")
    + """\
    s4 -> s5 [NAME] tag=verb;
"""
    + get_argument_references_part("s5", ["s7"], tagname="from_argument_name", prefix="from")
    + """\
    s7 -> s8 [NAME] tag=prepos;
"""
    + get_argument_references_part("s8", ["s10"], tagname="to_argument_name", prefix="to")
    + """\
    s10 -> s11 [WITH_KW];
    s11 -> s12 [SUB_CLAUSES_KW] tag=with_subclause;

    end accept=transform_main_with_subs;
    s12 -> end [NL_TK];
    s12 -> end [EOF_TK];
"""
)


def _process_transform_with_subs(
    tags: typing.TokensDict, _accept: str, builder: "AstBuilder"
) -> None:
    label_tok = tags["label"][0]
    doesaux_tok = tags["doesaux"][0]
    verb_tok = tags["verb"][0]
    in_flows = [components.Flow(name_tok) for name_tok in tags["from_argument_name"]]
    prepos_tok = tags["prepos"][0]
    out_flows = [components.Flow(name_tok) for name_tok in tags["to_argument_name"]]
    transform = components.Transformation(
        label_tok, doesaux_tok, verb_tok, in_flows, prepos_tok, out_flows
    )

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.add_transform(transform)


def _process_transform_subclause(
    tags: typing.TokensDict, _accept: str, builder: "AstBuilder"
) -> None:
    sub = sub_clause.decode_subclause(tags)
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.add_transform_subclause(sub)


MACHINES: typing.MachineTripletList = [
    ("TRANSFORM_HEADER_MACHINE", _TRANSFORM_HEADER_SPEC, _process_transform_header),
    (
        "TRANSFORM_MAIN_NO_SUBS_MACHINE",
        _TRANSFORM_MAIN_NO_SUBS_SPEC,
        _process_transform_no_subs,
    ),
    (
        "TRANSFORM_MAIN_WITH_SUBS_MACHINE",
        _TRANSFORM_MAIN_WITH_SUBS_SPEC,
        _process_transform_with_subs,
    ),
    (
        "TRANSFORM_SUB_CLAUSE_MACHINE",
        sub_clause.SUB_CLAUSE_SPEC,
        _process_transform_subclause,
    ),
]



"""Matrix based views on an ESL specification."""
from typing import List, Optional

import ragraph.plot
from plotly import graph_objs as go
from ragraph.colors import get_categorical
from ragraph.graph import Graph, GraphView
from ragraph.node import Node

from raesl.plot import view_funcs
from raesl.plot.generic import Style


def mdm(
    graph: Graph,
    node_kinds: Optional[List[str]] = None,
    edge_kinds: Optional[List[str]] = None,
    edge_labels: Optional[List[str]] = None,
    edge_weights: Optional[List[str]] = None,
    lead_components: Optional[List[Node]] = None,
    depth: Optional[int] = 2,
    style: Style = Style(),
) -> go.Figure:
    """Create a Multi-Domain Matrix plot using Plotly.

    Arguments:
        node_kinds: The node kinds to display.
        edge_kinds: The edge kinds to display.
        edge_labels: The edge labels to display.
        edge_weights: The edge weights to display.
        lead_components: The lead components to be used in node selection.
        depth: The depth up to which components and related nodes must be included.
        style: RaESL style options.

    Returns:
       Plotly :obj:`go.Figure` object of the Multi-Domain Matrix.
    """
    view = GraphView(
        graph,
        view_func=view_funcs.multi_domain,
        view_kwargs={
            "node_kinds": node_kinds,
            "edge_kinds": edge_kinds,
            "edge_labels": edge_labels,
            "edge_weights": edge_weights,
            "lead_components": lead_components,
            "depth": depth,
        },
    )

    style = style.ragraph
    if style.piemap.display == "labels":
        if not style.piemap.fields:
            style.piemap.fields = edge_labels

        if not style.palettes.get("fields"):
            style.palettes = {
                "fields": {
                    field: color
                    for field, color in zip(
                        graph.edge_labels,
                        get_categorical(n_colors=len(graph.edge_labels)),
                    )
                }
            }

    elif style.piemap.display == "kinds":
        if not style.piemap.fields:
            style.piemap.fields = edge_kinds
    elif style.piemap.display in ["weights", "weight labels"]:
        if not style.piemap.fields:
            style.piemap.fields = edge_weights

    return ragraph.plot.mdm(leafs=view.nodes, edges=view.edges, show=False, style=style, sort=False)



"""Graphviz diagrams of an ESL specification."""
from typing import Dict, List, Tuple

from graphviz import Digraph
from raesl.plot import utils, view_funcs
from raesl.plot.generic import Style
from ragraph.graph import Graph, GraphView
from ragraph.node import Node


def hierarchy_diagram(
    graph: Graph, roots: List[Node], levels: int, style: Style = Style()
) -> Digraph:
    """Draw a hierarchical decomposition tree using Graphviz.

    Arguments:
        graph: Instantiated ESL graph.
        roots: List of root nodes for which a tree must be drawn.
        levels: Number of levels to include in the tree.
        style: RaESL style options.

    Returns:
        Graphviz Digraph object of the decomposition tree.
    """
    view = GraphView(
        graph=graph,
        view_func=view_funcs.hierarchy,
        view_kwargs={"roots": roots, "levels": levels},
    )

    dot = _draw_migrated_graph(view, view.nodes, style)

    return dot


def functional_dependency_diagram(
    graph: Graph, root: Node, levels: int, style: Style = Style()
) -> Digraph:
    """Draw a functional dependency diagram using Graphviz.

    Arguments:
        graph: Instantiated ESL graph.
        root: Root node for which the dependency structure must be drawn.
        levels: Number of levels to include in the tree.
        style: RaESL style options.

    Returns:
        Graphviz Digraph object of the functional dependency diagram.
    """
    if root.kind != "component":
        raise ValueError(
            "Root node is of kind '{}' while it must be of kind 'component'.".format(root.kind)
        )

    view = GraphView(
        graph,
        view_func=view_funcs.functional_dependency,
        view_kwargs={"root": root, "levels": levels},
    )

    if style.diagram.show_hierarchy:
        roots = [root]
    else:
        roots = [n for n in view.nodes if n.kind == "component"]

    dot = _draw_migrated_graph(view, roots, style)

    return dot


def functional_context_diagram(
    graph: Graph, root: Node, degree: int = 1, style: Style = Style()
) -> Digraph:
    """Draw a functional context diagram using Graphviz.

    Arguments:
        graph: Instantiated ESL graph.
        root: Root node for which the functional context is drawn.
        degree: The degree up to which neighbors must be collected (neighbors of
            neighbors). Defaults to 1.
        style: RaESL style options.

    Returns:
        Graphviz Digraph object of the functional context diagram.
    """
    if root.kind != "component":
        raise ValueError(
            "Root node is of kind '{}' while it must be of kind 'component'.".format(root.kind)
        )

    view = GraphView(
        graph,
        view_func=view_funcs.functional_context,
        view_kwargs={"root": root, "degree": degree, "style": style},
    )

    components = [n for n in view.nodes if n.kind == "component"]
    if style.diagram.show_root_children or style.diagram.show_neighbor_children:
        roots = [c for c in components if c.depth <= root.depth]
        roots.extend([c.parent for c in components if c.parent.depth == root.depth])
    else:
        roots = components

    dot = _draw_migrated_graph(view, roots, style=style)

    return dot


def function_chain_diagram(
    graph: Graph,
    start_points: List[Tuple[Node, List[Node]]],
    end_points: List[Tuple[Node, List[Node]]],
    levels: int = 1,
    style: Style = Style(),
) -> Digraph:
    """Draw a function chain diagram using Graphviz.

    Arguments:
        graph: Instantiated ESL graph.
        start_points: List of tuples that contain the component node and list of
            function nodes that serve as the starting point of the function chains.
        end_points: List of tuples that contain the component node and list of
            function nodes that serve as the end point of the function chains.
        levels: Number of levels to decompose intermediate components into.
            This number is relative to the depth of the start nodes and end nodes.
        style: RaESL style options.

    Returns:
        Graphviz Digraph object of the function chain diagram.
    """
    utils.check_start_and_end_points(start_points=start_points, end_points=end_points)

    view = GraphView(
        graph,
        view_func=view_funcs.function_chain,
        view_kwargs=dict(
            start_points=start_points, end_points=end_points, levels=levels, style=style
        ),
    )

    components = [n for n in view.nodes if n.kind == "component"]
    if style.diagram.show_hierarchy:
        ancestors = set([a for c in components for a in c.ancestors])
        roots = [n for n in ancestors if not n.parent]
    else:
        roots = components

    dot = _draw_migrated_graph(view, roots, style)

    return dot


def function_traceability_diagram(
    graph: Graph, root: Node, levels: int, style: Style = Style()
) -> Digraph:
    """Draw a functional traceability diagram using Graphviz.

    Arguments:
        graph: Instantiated ESL graph.
        root: Node that serves a the root of the traceability tree. Must be a
            transformation specification.
        levels: Number of levels to go down into the traceability tree.
        style: RaESL Style options.

    Returns
        Graphviz Digraph object of the functional traceability diagram.
    """

    # Input checks.
    if root.kind != "function_spec":
        raise ValueError(
            "Root node must be of kind 'function_spec' not of kind '{}'".format(root.kind)
        )

    if root.annotations.esl_info["sub_kind"] != "transformation":
        raise ValueError(
            """Root node must be of sub-kind 'transformation'
            not of sub-kind '{}'""".format(
                root.annotations.esl_info["sub_kind"]
            )
        )

    view = GraphView(
        graph,
        view_func=view_funcs.traceability,
        view_kwargs=dict(root=root, levels=levels, style=style),
    )

    components = [n for n in view.nodes if n.kind == "component"]
    dot = _draw_migrated_graph(view, components, style)

    return dot


def _draw_migrated_graph(
    view: GraphView,
    roots: List[Node],
    style: Style,
) -> Digraph:
    """Get a correctly migrated :obj:`graphviz.Digraph` of a :obj:`Graph`.

    Arguments:
        view: GraphView using one of the ESL filters below.
        roots: List of root nodes.
        style: RaESL style options.
    """
    digraph = Digraph("G", comment=view.name, **style.diagram.digraph)
    digraph.graph_attr["rankdir"] = style.diagram["orientation"]

    trans, goals = utils.get_component_function_dicts(view)
    cmps = set([n for n in view.nodes if n.kind == "component"])
    leafs = {c.name: c for c in cmps if not c.children or not set(c.children).intersection(cmps)}

    for r in roots:
        if r.name in leafs:
            _add_component(digraph, r, trans, goals, style)
        else:
            migrated_goals = []
            for a in r.ancestors:
                migrated_goals.extend(goals.get(a.name, []))

            for goal in migrated_goals:
                _add_goal(digraph, goal, style)

            _add_cluster(digraph, r, leafs, trans, goals, style)

    for e in view.edges:
        src = _get_display_name(e.source, trans, leafs)
        trg = _get_display_name(e.target, trans, leafs)
        digraph.edge(src, trg, style=style.diagram.edge_styles[e.kind])

    return digraph


def _add_component(
    digraph: Digraph,
    node: Node,
    trans: Dict[str, List[Node]],
    goals: Dict[str, List[Node]],
    style: Style,
):
    """Adding a component to a Digraph object. Associated goals
    and transformations are added as well.

    Arguments:
        digraph: Digraph object to add the transformation to.
        node: The component to be added.
        trans: Dictionary of component node names to a lists of transformation
           specifications.
        goals: Dictionary of component node names to lists of goal specifications.
        style: RaESL style options.
    """
    if trans[node.name]:
        with digraph.subgraph(name="cluster_" + node.name) as sg:
            sg.attr(label=node.name)
            for t in trans[node.name]:
                _add_transformation(sg, t, style)
    else:
        digraph.node(
            node.name,
            label=node.name.split(".")[-1].replace("_", " "),
            shape=style.diagram.node_shapes["component"],
        )

    for goal in goals[node.name]:
        _add_goal(digraph, goal, style)


def _add_cluster(
    digraph: Digraph,
    node: Node,
    leafs: Dict[str, Node],
    trans: Dict[str, List[Node]],
    goals: Dict[str, List[Node]],
    style: Style,
):
    """Adding a cluster to a Digraph object. Associated goals are added as well.
    Children are added recursively.

    Arguments:
        digraph: Digraph object to add the transformation to.
        node: The component to be added.
        leafs: Dictionary of leaf node names to leaf nodes.
        trans: Dictionary of component node names to a lists of transformation
           specifications.
        goals: Dictionary of component node names to lists of goal specifications.
        style: RaESL style options.
    """

    with digraph.subgraph(name="cluster_" + node.name) as sd:
        sd.attr(label=node.name)
        for child in node.children:
            if child.name in leafs:
                _add_component(sd, child, trans, goals, style)
            else:
                _add_cluster(sd, child, leafs, trans, goals, style)

    for goal in goals[node.name]:
        _add_goal(digraph, goal, style=style)


def _add_goal(digraph: Digraph, node: Node, style: Style):
    """Adding a goal specification to a Digraph object.

    Arguments:
        digraph: Digraph object to add the goal specification to.
        node: The Node corresponding to the goal specification to be added.
        style: RaESL style options.
    """
    if style.diagram.list_variables:
        label = "\n - ".join(
            [node.name + "\n Variables:"] + node.annotations.esl_info["body"]["variables"]
        )
        digraph.node(
            name=node.name,
            label=label,
            shape=style.diagram.node_shapes["goal"],
        )
    else:
        digraph.node(name=node.name, shape=style.diagram.node_shapes["goal"])


def _add_transformation(digraph: Digraph, trans: Node, style: Style):
    """Adding a transformation specification to a Digraph object.

    Arguments:
        digraph: Digraph object to add the transformation to.
        trans: The Node corresponding to the transformation specification to be added.
        style: RaESL style options.
    """
    if style.diagram.list_variables:
        inputs = "\n -".join(
            ["Input variables:"] + trans.annotations.esl_info["body"]["input_variables"]
        )
        outputs = "\n -".join(
            ["Output variables:"] + trans.annotations.esl_info["body"]["output_variables"]
        )
        label = "\n".join([trans.name, inputs, outputs])
        digraph.node(
            name=trans.name,
            label=label,
            shape=style.diagram.node_shapes["transformation"],
        )
    else:
        digraph.node(trans.name, shape=style.diagram.node_shapes["transformation"])


def _get_display_name(node: Node, trans: Dict[str, List[str]], leafs: Dict[str, Node]) -> str:
    """Get the display name for a Node in the diagram.

    Arguments:
        node: Node to get the display name for.
        trans: Dictionary of component node names to a list of transformation specs.
        leafs: Dictionary of leaf node names to leaf nodes.

    Returns:
        Display name of the Node.

    Note:
        Clusters are prepended with ``cluster_``.
    """
    if _is_clusternode(node, trans, leafs):
        return "cluster_" + node.name

    return node.name


def _is_clusternode(node: Node, trans: Dict[str, List[str]], leafs: Dict[str, Node]):
    """Check whether a Node represents a clusternode in the diagram.

    Arguments:
        node: Node to get the display name for.
        trans: Dictionary of component node names to a list of transformation specs.
        leafs: Dictionary of leaf node names to leaf nodes.

    Returns:
        Whether the Node represents a clusternode in the diagram.
    """
    if node.kind != "component":
        return False

    if node.name not in leafs:
        return True

    if trans.get(node.name):
        return True

    return False



"""Line matcher for sub-clause lines."""
import re
import string
from typing import TYPE_CHECKING, Dict, List, Optional, cast

from raesl.compile.ast import components, exprs
from raesl.compile.machine_files.machine_parts import get_disjunctive_comparison_part
from raesl.compile.machine_files.utils import get_one, get_optional

if TYPE_CHECKING:
    from raesl.compile.scanner import Token


SUB_CLAUSE_SPEC = (
    """
sub_clause:
    start initial;
    start -> s1 [STAR_TK];
    s1 -> s2 [NAME] tag=subclause_label;
    s2 -> s3 [COLON_TK];
"""
    + get_disjunctive_comparison_part("s3", ["s4", "s5"])
    + """\
    finish accept=sub_clause;
    s4 -> finish [NL_TK];
    s4 -> finish [EOF_TK];
    s5 -> finish [NL_TK];
    s5 -> finish [EOF_TK];
"""
)


def guess_is_var(varvalue: "Token") -> bool:
    """Guess whether the provided token is a variable or a value.
    (Answer: If it is not "t.b.d." and starts with a letter it's a variable.)
    """
    text = varvalue.tok_text
    if re.fullmatch("[tT]\\.[Bb]\\.[dD]\\.", text):
        # Some form of TBD is not a variable.
        return False

    first = text[0]
    return first in string.ascii_lowercase or first in string.ascii_uppercase


def decode_disjunctive_comparisons(tags: Dict[str, List["Token"]]) -> exprs.Expression:
    """Decode tags of a matched 'machine_parts.get_disjunctive_comparison_part' part
    to an disjunction with comparisons.

    Arguments:
        tags: Extracted data from a match of the machine defined in
            'machine_parts.get_disjunctive_comparison_part'.

    Returns:
        The expression equivalent to the matched text.
    """
    split_offsets = (
        cast(List[Optional[int]], [None]) + [tok.offset for tok in tags.get("or", [])] + [None]
    )
    equations: List[exprs.Expression] = []

    unit_tags = tags.get("unit", [])
    compare_tags = tags.get("compare_op", [])
    for index in range(1, len(split_offsets)):
        start_offset = split_offsets[index - 1]
        end_offset = split_offsets[index]

        first_var = get_one(tags["first_var"], start_offset, end_offset)
        lhs = exprs.VariableValue(first_var)

        is_aux = get_one(tags["is_aux"], start_offset, end_offset)
        is_constraint = is_aux.tok_text == "is"

        comp: exprs.Comparison
        compare_op = get_optional(compare_tags, start_offset, end_offset)
        if compare_op is not None:
            varvalue = get_one(tags["varvalue"], start_offset, end_offset)
            unit = get_optional(unit_tags, start_offset, end_offset)

            rhs: exprs.DataValue
            if unit is None and guess_is_var(varvalue):
                rhs = exprs.VariableValue(varvalue)
            else:
                rhs = exprs.Value(varvalue, unit)

            comp = exprs.RelationComparison(is_constraint, lhs, is_aux, compare_op, rhs)
            equations.append(comp)

        else:
            objective = get_one(tags["objective"], start_offset, end_offset)

            comp = exprs.ObjectiveComparison(lhs, is_aux, objective.tok_type == "MAXIMIZED_KW")
            equations.append(comp)

    if len(equations) == 1:
        return equations[0]
    else:
        return exprs.Disjunction(equations)


def decode_subclause(tags: Dict[str, List["Token"]]) -> components.SubClause:
    """Decode tags of a matched subclauses line to one or more disjunctive equations.

    Arguments:
        tags: Extracted data from a match of the machine defined in SUB_CLAUSE_SPEC.

    Returns:
        The found subclause.
    """
    label = tags["subclause_label"][0]
    condition = decode_disjunctive_comparisons(tags)
    return components.SubClause(label, condition)



"""Line matcher state machines for designs."""
from typing import TYPE_CHECKING

from raesl.compile.ast import components
from raesl.compile.machine_files import sub_clause, typing
from raesl.compile.machine_files.machine_parts import get_disjunctive_comparison_part

if TYPE_CHECKING:
    from raesl.compile.typechecking.ast_builder import AstBuilder


_DESIGN_HEADER_SPEC = """
design_header:
    start initial;
    start -> s1 [DESIGN_REQUIREMENT_KW] tag=design;
    start -> s1 [DESIGN_CONSTRAINT_KW] tag=design;

    end accept=design_header;
    s1 -> end [NL_TK];
    s1 -> end [EOF_TK];
"""


def _process_design_header(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    kind = tags["design"][0]
    builder.notify_new_section(kind, False)

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.new_design_header(kind)


_DESIGN_WITH_SUBS_SPEC = (
    """
design_clause_with_sub:
    start initial;
    start -> s1 [NAME] tag=label;
    s1 -> s2 [COLON_TK];
"""
    + get_disjunctive_comparison_part("s2", ["s3", "s4"])
    + """\
    s3 -> s5 [WITH_KW];
    s4 -> s5 [WITH_KW];
    s5 -> s6 [SUB_CLAUSES_KW] tag=with_subclause;

    finish accept=design_with_subs;
    s6 -> finish [NL_TK];
    s6 -> finish [EOF_TK];
"""
)


def _process_design_line(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    label = tags["label"][0]
    condition = sub_clause.decode_disjunctive_comparisons(tags)
    design = components.Design(label, condition)

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.design_line(design)


_DESIGN_NO_SUBS_SPEC = (
    """
design_clause_no_sub:
    start initial;
    start -> s1 [NAME] tag=label;
    s1 -> s2 [COLON_TK];
"""
    + get_disjunctive_comparison_part("s2", ["s3", "s4"])
    + """\
    finish accept=design_no_subs;
    s3 -> finish [NL_TK];
    s3 -> finish [EOF_TK];
    s4 -> finish [NL_TK];
    s4 -> finish [EOF_TK];
"""
)

# _DESIGN_NO_SUBS_SPEC also uses _process_design_line.


def _process_design_subclause(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    sub = sub_clause.decode_subclause(tags)

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.add_design_subclause(sub)


MACHINES: typing.MachineTripletList = [
    ("DESIGN_HEADER_MACHINE", _DESIGN_HEADER_SPEC, _process_design_header),
    ("DESIGN_WITH_SUBS_MACHINE", _DESIGN_WITH_SUBS_SPEC, _process_design_line),
    ("DESIGN_NO_SUBS_MACHINE", _DESIGN_NO_SUBS_SPEC, _process_design_line),
    (
        "DESIGN_SUB_CLAUSE_MACHINE",
        sub_clause.SUB_CLAUSE_SPEC,
        _process_design_subclause,
    ),
]



"""Module to generate text lines that describe ESL objects."""

from typing import Dict, Generator, Iterable, List, Optional

from raesl.doc.locales import _, hookspec, pm
from ragraph.graph import Graph
from ragraph.node import Node

LineGen = Generator[str, None, None]


class Hookspecs:
    @hookspec(firstresult=True)
    def linguistic_enumeration(items: List[str]) -> str:
        """Get a natural language enumeration of items."""

    @hookspec(firstresult=True)
    def linguistic_options(items: List[str]) -> str:
        """Get a natural language enumeration of options."""

    @hookspec(firstresult=True)
    def function_spec_node(node: Node, graph: Graph) -> LineGen:
        """Yield the function spec in natural language."""

    @hookspec(firstresult=True)
    def design_spec_node(node: Node, graph: Graph) -> LineGen:
        """Yield the design spec in natural language."""

    @hookspec(firstresult=True)
    def behavior_spec_node(node: Node, graph: Graph) -> LineGen:
        """Yield the behavior spec in natural language."""

    @hookspec(firstresult=True)
    def need_node(node: Node, graph: Graph) -> LineGen:
        """Yield the need spec in natural language."""


pm.add_hookspecs(Hookspecs)
hook = pm.hook


def hs(h: int):
    """Get header pound (#) signs."""
    return h * "#"


def header(h: int, text: str, capitalize: bool = True) -> str:
    """Get a header with surrounding whitespace and optional capitalization."""
    if capitalize:
        text = cap(text)
    return "\n\n{} {}\n\n".format(hs(h), text)


def boldhead(text: str, capitalize: bool = True, newlines: bool = True) -> str:
    """Get a bold header (without numbering) with surrounding whitespace."""
    if capitalize:
        text = cap(text)

    text = bold(text)

    if newlines:
        text = "\n{}\n".format(text)

    return text


def emph(text: str) -> str:
    """Return text as emphasized (Markdown)."""
    return "*{}*".format(text.strip())


def cap(text: str) -> str:
    """Capitalize first char of text."""
    return text[0].upper() + text[1:]


def bold(text: str) -> str:
    """Return text as bold (markdown)."""
    return "**{}**".format(text.strip())


def snt(text: str) -> str:
    """Text to a sentence. Capitalizes first character and adds a period."""
    if not text.endswith("."):
        text += "."
    return cap(text)


def ordered(items: Iterable[str], indent: int = 0) -> LineGen:
    """Generate an ordered Markdown list."""
    yield ""
    for item in items:
        if type(item) == str:
            yield "{}1. {}".format(indent * " ", item)  # Markdown handles numbering.
        else:
            yield from ordered(item, indent=2)
    if indent == 0:
        yield "\n<!-- end of list -->\n"  # Ensures list is ended here.


def unordered(items: Iterable[str], indent: int = 0) -> LineGen:
    """Generate an unordered Markdown list."""
    yield ""
    for item in items:
        if type(item) == str:
            yield "{}* {}".format(indent * " ", item)
        else:
            yield from unordered(item, indent=2)
    if indent == 0:
        yield "\n<!-- end of list -->\n"  # Ensures list is ended here.


def image(
    path: str,
    caption: Optional[str] = None,
    label: Optional[str] = None,
    sizing: Optional[str] = None,
) -> str:
    """Get Pandoc Markdown for an image."""
    line = "\n\n![{}]({})".format(caption, path)
    attrs = ""
    if label:
        attrs += "#fig:{}".format(label)
    if sizing:
        attrs += " {}".format(sizing)
    if attrs:
        line += "{" + attrs + "}"
    return line + "\n\n"


def node_path(path: str, italic: bool = False, arrows: bool = True, skip: str = "world"):
    """Get a friendly representation of a node path."""
    skips = skip.split(".")
    for skip in skips:
        if path.startswith(skip):
            idx = len(skip) + 1
            path = path[idx:]
        else:
            break

    if arrows:
        path = path.replace(".", " &rarr; ")
    path = path.strip()
    if italic:
        return "*" + path + "*"
    else:
        return path


def bundle_path(
    path: str, root: str, italic: bool = False, arrows: bool = True, skip: str = "world"
):
    """Get a friendly representation of a bundle path."""
    path_parts = path.split("." + root + ".")

    if path_parts[0] in skip:
        path = root + "." + path_parts[-1]
    else:
        path = (
            node_path(path=path_parts[0] + ".", italic=False, arrows=arrows, skip=skip)
            + " "
            + root
            + "."
            + path_parts[-1]
        )

    return "*" + path + "*" if italic else path


def var_path(v: Node, italic: bool = False, arrows: bool = True, skip: str = "world"):
    """Get a friendly representation of a variable path."""
    if v.annotations.esl_info.get("bundle_root_name"):
        return bundle_path(
            path=v.name,
            root=v.annotations.esl_info["bundle_root_name"],
            arrows=arrows,
            skip=skip,
            italic=italic,
        )
    else:
        return node_path(v.name, italic=italic, arrows=arrows, skip=skip)


def var_clarification(bvars: List[Node]):
    """Yield variable clarification section."""
    if len(bvars) == 1:
        yield "\n"
        yield snt(
            _("where the full name of variable {} is {}").format(
                bvars[0].name.split(".")[-1],
                bundle_path(
                    path=bvars[0].name,
                    root=bvars[0].annotations.esl_info["bundle_root_name"],
                ),
            )
        )
    elif len(bvars) > 1:
        yield "\n"
        yield cap(_("where, respectively:"))
        yield from unordered(
            [
                _("variable {} has full name {}").format(
                    v.name.split(".")[-1],
                    bundle_path(
                        path=v.name,
                        root=v.annotations.esl_info["bundle_root_name"],
                    ),
                )
                for v in bvars
            ]
        )


def bundle_clarification(brvdict: Dict[str, List[Node]]):
    """Yield bundle clarification section"""
    yield "\n"
    verbs = _("is a bundle") if len(brvdict) == 1 else _("are bundles")
    yield cap(
        _("where {} {} of which the following variables are used:").format(
            pm.hook.linguistic_enumeration(items=list(brvdict.keys())), verbs
        )
    )
    yield "\n"
    yield from unordered(
        [bundle_path(path=vname, root=key) for key in brvdict for vname in brvdict[key]]
    )


def component_node(node: Node, h: int) -> LineGen:
    """Yield component section"""
    nameparts = node.name.split(".")
    yield header(h, "{}".format(nameparts[-1]))
    yield snt(_("this section describes **{}**".format(nameparts[-1])))
    if node.parent.name != "world":
        yield snt(_("this component is a sub-component of {}").format(node_path(node.parent.name)))

    comments = node.annotations.esl_info.get("comments", [])
    if comments:
        yield "\n"
        yield boldhead(_("comments")).replace("\n", "")
        yield "\n"
        yield from comments

    for key, comments in node.annotations.esl_info.get("tagged_comments", {}).items():
        yield "\n"
        yield boldhead(key).replace("\n", "")
        yield "\n"
        yield from comments


kind_mapping = {
    "component": component_node,
    "function_spec": hook.function_spec_node,
    "design_spec": hook.design_spec_node,
    "behavior_spec": hook.behavior_spec_node,
    "need": hook.need_node,
}


def lines(node: Node, **kwargs) -> LineGen:
    """Yield lines that describes a Node.

    Arguments:
        node: Node to generate the lines from.
    """

    genfunc = kind_mapping[node.kind]
    yield from genfunc(node=node, **kwargs)


def get_design_rule_line_vars(rules: List[Dict[str, str]], g: Graph):
    """Get variables that are used within a design rule line"""
    vrs = [
        g[r["subject"]]
        for r in rules
        if g[r["subject"]].annotations.esl_info.get("bundle_root_name")
    ]

    vrs += [
        g[r["bound"]["value"]]
        for r in rules
        if g.node_dict.get(r["bound"]["value"], None)
        and g[r["bound"]["value"]].annotations.esl_info.get("bundle_root_name")
    ]

    return vrs



"""Typing support for the machines data."""
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Tuple

if TYPE_CHECKING:
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.ast_builder import AstBuilder

# Function signature of the processing function.
TokensDict = Dict[str, List["Token"]]
ProcessingFunc = Callable[[TokensDict, str, "AstBuilder"], None]

# Triplet in MACHINES variable.
MachineTriplet = Tuple[str, str, Optional[ProcessingFunc]]

# Type of the MACHINES variable.
MachineTripletList = List[MachineTriplet]



"""Line matcher state machines for verb and pre-position definitions."""
from typing import TYPE_CHECKING

from raesl.compile.machine_files import typing

if TYPE_CHECKING:
    from raesl.compile.typechecking.ast_builder import AstBuilder


_DEFINE_VERB_SPEC = """
define_verb:
    start initial;
    start -> s1 [DEFINE_KW] tag=define;
    s1 -> s2 [VERB_KW];

    end accept=define_verb;
    s2 -> end [NL_TK];
    s2 -> end [EOF_TK];
"""


def _process_define_verb(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    builder.notify_new_section(tags["define"][0], True)


_VERB_PREPOS_SPEC = """
verb_prepos:
    start initial;
    start -> s1 [NAME] tag=verb;
    s1 -> s2 [NAME] tag=prepos;

    end accept=verb_is_verb;
    s2 -> end [NL_TK];
    s2 -> end [EOF_TK];
"""


def _process_verb_def(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    verb = tags["verb"][0]
    prepos = tags["prepos"][0]
    builder.add_verbdef(verb, prepos)


MACHINES: typing.MachineTripletList = [
    ("DEFINE_VERB_MACHINE", _DEFINE_VERB_SPEC, _process_define_verb),
    ("VERB_PREPOS_MACHINE", _VERB_PREPOS_SPEC, _process_verb_def),
]



"""Generic utility functions."""
import os
import urllib.parse
from pathlib import Path
from typing import Dict, Generator, List, Optional, Set, Tuple, Union

from ragraph.graph import Graph
from ragraph.node import Node

from raesl import logger
from raesl.types import Location, Position, Range


def get_esl_paths(*paths: Union[str, Path]) -> List[Path]:
    """Get a sorted list of ESL file paths from multiple file or directory paths."""
    if not paths:
        raise ValueError("No paths were specified.")

    result: Set[Path] = set()

    pathlist = list(paths)
    for path in pathlist:
        logger.debug(f"Resolving '{path}'...")
        if isinstance(path, list):
            pathlist.extend(path)
            continue
        p = Path(path)

        if not p.exists():
            logger.info(f"Skipped '{p}' as it does not exist.")
            continue

        if p.is_dir():
            result.update(
                p for p in p.glob("**/*.esl") if not any(part.startswith(".") for part in p.parts)
            )

        if p.is_file():
            result.add(p)

    if not result:
        raise ValueError("No ESL files found.")

    return sorted(result)


def check_output_path(fpath: Union[str, Path], force: bool) -> Path:
    """Check output filepath versus force overwrite status."""
    p = Path(fpath)

    if p.exists() and not force:
        raise ValueError(f"Path {p} already exists and force overwrite isn't set.")

    if p.is_dir():
        raise ValueError(f"Path {p} is a directory.")

    return p


def get_location(
    uri: str = "Unknown",
    start_line: int = 0,
    start_character: int = 0,
    end_line: Optional[int] = None,
    end_character: Optional[int] = None,
) -> Location:
    """Generation utility to quickly drum up a location.

    Arguments:
        uri: Location uri.
        start_line: Location's range start line.
        start_character: Location's range start offset.
        end_line: Optional Location's range end line (otherwise identical to start.)
        end_character: Optional Location's range end offset (otherwise identical to
            start.)

    Returns:
        Newly created location instance.
    """
    end_line = start_line if end_line is None else end_line
    end_character = start_character if end_character is None else end_character

    return Location(
        uri,
        Range(Position(start_line, start_character), Position(end_line, end_character)),
    )


def cleanup_path(path: Union[str, Path]) -> Path:
    """Cleanup pathname for some typical mistakes."""
    p = str(path)
    result = uri_to_path(p)
    return result


def uri_to_path(uri: str) -> Path:
    """Convert a file URI to a regular path."""
    parsed = urllib.parse.unquote(urllib.parse.urlparse(uri).path)

    if os.name == "nt" and parsed.startswith("/"):
        parsed = parsed[1:]

    return Path(parsed)


def path_to_uri(path: Union[str, Path]) -> str:
    """Convert a path to a file URI."""
    if str(path).startswith("file:"):
        return str(path)
    return Path(path).resolve().as_uri()


def split_first_dot(name: str) -> Tuple[str, str, int]:
    """Split the provided name on the first dot if it exists, return both parts, and
    the length of the dot.
    """
    i = name.find(".")
    if i >= 0:
        return name[:i], name[i + 1 :], 1
    else:
        return name, "", 0


def get_first_namepart(name: str) -> str:
    """Return the name upto and excluding the first dot."""
    i = name.find(".")
    if i < 0:
        return name
    return name[:i]


def get_scoped_nodes(graph: Graph, scopes: Dict[str, Optional[int]]) -> List[Node]:
    """Get scoped nodes, being subtrees of the graph of varying depth.

    Arguments:
        graph: Graph data.
        scopes: Node names mapped to depths of the subtree to include. A depth of
            :obj:`None` includes the whole subtree starting at that node.

    Returns:
        List of nodes in all given scopes.
    """
    seen: Set[str] = set()
    nodes = []

    for name, depth in scopes.items():
        try:
            node = graph[name]
        except KeyError:
            raise KeyError(
                f"Node '{name}' does not exist in the graph. "
                "Make sure you provide the entire instantiated path "
                "(e.g. 'world.component.subcomponent')."
            )

        for candidate in yield_subtree(node, depth):
            if candidate.name in seen:
                continue
            nodes.append(candidate)

    return nodes


def yield_subtree(root: Node, depth: Optional[int]) -> Generator[Node, None, None]:
    """Yield nodes from a given subtree starting at Node and with given depth.

    Arguments:
        root: Root node of subtree.
        depth: Depth of subtree. If None, defaults to full depth.

    Yields:
        Nodes in the subtree.
    """
    yield root

    if depth is None or depth > 0:
        for c in root.children:
            yield from yield_subtree(c, None if depth is None else depth - 1)



"""Dutch locale overrides."""

from collections import defaultdict
from typing import Any, Dict, Generator, List

import raesl.doc.lines as lns
from raesl import logger
from raesl.doc.locales import hookimpl, pm
from ragraph.graph import Graph
from ragraph.node import Node

LineGen = Generator[str, None, None]


translations = {
    "system specification decomposition level": "systeemspecificatie decompositieniveau",
    "this chapters describes the system of interest at the first decomposition level. That is, it describes {} components which play a role within the environment in which the system of interest must operate and the (functional) interactions between those components": "dit hoofdstuk beschrijft het systeem op het eerste decompositie niveau. Dit niveau beschrijft {} componenten een die rol spelen in de wereld waarin het systeem dient te opereren en die (functionele) interacties daar tussen",
    "this chapters describes the system of interest at decomposition level {} and introduces one additional component": "dit hoofdstuk beschrijft het systeem op decompositieniveau {} en introdceert e\\'e\\'m extra component",
    "this chapters describes the system of interest at decomposition level {} and introduces {} additional component": "dit hoofdstuk beschrijft het systeem op decompositieniveau {} en introdceert {} extra component",
    "transformation function requirements": "transformatiefunctie-eisen",
    "transformation function constraints": "transformatiefunctie-randvoorwaarden",
    "qualitative design requirements": "kwalitatieve ontwerpeisen",
    "quantitative design requirements": "kwantitatieve ontwerpeisen",
    "quantitative design constraints": "kwantitatieve ontwerprandvoorwaarden",
    "sub-components": "sub-componenten",
    "sub-component": "sub-component",
    "goal function requirements": "doelfunctie-eisen",
    "goal function constraints": "doelfunctie-randvoorwaarden",
    "local component tree": "locale objectboom",
    "relations between variables": "relaties tussen variabelen",
    "dependency structure": "afhankelijkheidsstructuur",
    "this section describes **{}**": "deze sectie beschrijft **{}**",
    "this component is a sub-component of {}": "dit component is een sub-component van {}",
    "clarification": "toelichting",
    "of": "van",
    "world": "de wereld",
    "scope": "scope",
    "path": "pad",
    "the following sub-components are defined within {}:": "de volgende sub-componenten zijn gedefinieerd binnen {}:",
    "list of variables": "lijst van variabelen",
    "variable": "variabele",
    "type": "type",
    "definitions": "definities",
    "variable reference list": "referentielijst variabelen",
    "related to": "gerelateerd aan",
    "properties": "eigenschappen",
    "interval(s): ": "interval(len): ",
    "enumeration of ": "enumeratie van ",
    "units": "eenheden",
    "domain": "domein",
    " upper bound equals {}": " bovengrens gelijk aan {}",
    " and upper bound equals {}": " en bovengrens gelijk aan {}",
    "lower bound equals {}": "ondergrens gelijk aan {}",
    "the following properties are specified for {}": "de volgende eigenschappen zijn gedefinieerd voor {}",
    "relations and models": "relaties en modellen",
    "variables": "variabelen",
    "{} dependency matrix of decomposition level {}.": "{} raakvlakken matrix van decompositieniveau {}.",
    "in Figure \\ref{{{}}} the associated design-structure-matrix (DSM) is shown": "in Figuur \\ref{{{}}} is de bijbehorende dependency-structure-matrix (DSM) te zien",
    "the DSM shows the dependencies between the elements that are relevant to this decomposition level": "de DSM geeft de afhankelijkheden tussen de elementen die relevant zijn voor dit decompositieniveau.",
    "global qualitative design requirements": "systeembrede kwalitatieve ontwerpeisen",
    "global quantitative design requirements": "systeembrede kwantitatieve ontwerpeisen",
    "global quantitative design constraints": "systeembrede kwantitatieve ontwerprandvoorwaarden",
    "the MDM shows the dependencies between the components, the function specifications and the combinations thereof that are relevant to this decomposition level": "de MDM illustreert de afhankelijkheden tussen de componenten, tussen de functies en combinaties daarvan behorende bij dit decompositieniveau",
    "behavior requirements": "gedragseisen",
    "behavior constraints": "gedragsrandvoorwaarden",
    "{} is composed of the following sub-components:": "{} bestaat uit de volgende subcomponenten:",
    "this is the first decomposition level which indicates which components play a role in the environment in which the system at hand must operate": "dit is het eerste decompositieniveau waarin de elementen worden beschreven waarmee het systeem interactie heeft of dient te hebben.",
    "subordinate function specifications": "onderliggende functie specificaties",
    "external models": "externe modellen",
    "model definition name": "naam van de model definitie",
    "required variables": "input variabelen",
    "returned variables": "output variabelen",
    "related variables": "gerelateerde variabelen",
    "this goal function requirement automatically migrated from {}": "deze doelfunctie-eis is automatisch ge-migreerd van {}",
    "this goal function constraint automatically migrated from {}": "deze doelfunctie-randvoorwaarde is automatisch ge-migreerd van {}",
    "system-wide qualitative and quantitative specifications": "systeembrede kwalitatieve and kwantitatieve ontwerpeisen",
    "this chapter lists all qualitative and quantitative design specifications that cannot be linked to a component": "dit hoofdstuk bevat alle kwalitatieve and kwantitatieve ontwerpeisen die niet aan een component zijn gerelateerd",
    "is a bundle": "een bundel is",
    "are bundles": "bundles zijn",
    "where {} {} of which the following variables are used:": "waar {} {} waaruit de volgende variabelen worden gebruikt:",
    "where the full name of variable {} is {}": "waar de variabel {} de volledige naam {} heeft",
    "where, respectively:": "waar, respectivelijk",
    "variable {} has full name {}": "variabel {} de volledige naam {} heeft",
}


@hookimpl
def gettext(key: str):
    """Get translated string."""
    if key in translations:
        return translations[key]
    logger.debug("Cannot find a translation for '{}' in {}.".format(key, __file__))
    return key


_ = gettext


@hookimpl
def linguistic_enumeration(items: List[str]):
    n = len(items)
    if n <= 2:
        return " en ".join(items)
    return ", ".join(items[:-1]) + " en " + items[-1]


@hookimpl
def linguistic_options(items: List[str]):
    n = len(items)
    if n <= 2:
        return " of ".join(items)
    return ", ".join(items[:-1]) + " of " + items[-1]


def subclause_line(s: Dict[str, Any]):
    """Yield subclause line."""

    yield " of ".join([subclause_rule_line(r=r) for r in s["body"]])


def is_number(v: Any) -> bool:
    try:
        float(v)
        return True
    except ValueError:
        return False


drule_auxmap = {
    "must be": ("dient", "te zijn"),
    "shall be": ("zal", "zijn"),
    "should be": ("behoort", "te zijn"),
    "could be": ("zou", "kunnen zijn"),
    "won't": ("zal", "zijn"),
    "is": ("is", ""),
}

cmap = {
    "smaller than": "kleiner dan",
    "at most": "hoogstens",
    "equal to": "gelijk aan",
    "at least": "tenminste",
    "greater than": "groter dan",
    "approximately": "ongeveer",
}


def make_predicate(verb: str) -> str:
    """Check if a "splitsend werkwoord" has been used and return predicate.

    Argument:
      verb: the used verb

    Returns
      Predicate of the sentence.
    """

    preps = [
        "aan",
        "in",
        "op",
        "om",
        "na",
        "tegen",
        "tussen",
        "uit",
        "bij",
        "mee",
        "af",
        "mee",
        "terug",
    ]

    for prep in preps:
        if verb.startswith(prep):
            return prep + " te " + verb[len(prep) :]

    return " te " + verb


def design_rule_line(r: Dict[str, str]):
    s = r["subject"].split(".")[-1]
    aux = drule_auxmap.get(r["auxiliary"])
    comparison = cmap.get(r["comparison"])

    if is_number(v=r["bound"]["value"]):
        value = r["bound"]["value"]
    elif r["bound"]["value"] == "t.b.d.":
        value = "n.t.b."
    else:
        value = r["bound"]["value"].split(".")[-1]
    unit = ""
    if r["bound"]["unit"]:
        unit = r["bound"]["unit"]

    if r["auxiliary"] == "won't":
        return "{} {} nimmer {} {} {} {}".format(s, aux[0], comparison, value, unit, aux[1])
    else:
        return "{} {} {} {} {} {}".format(s, aux[0], comparison, value, unit, aux[1])


def then_clause_line(s: Dict[str, Any]):
    """Yield subclause line."""

    yield " of ".join([then_clause_rule_line(r=r) for r in s["body"]])


def then_clause_rule_line(r: Dict[str, str]):
    s = r["subject"].split(".")[-1]
    aux = drule_auxmap.get(r["auxiliary"])
    comparison = cmap.get(r["comparison"])

    if is_number(v=r["bound"]["value"]):
        value = r["bound"]["value"]
    elif r["bound"]["value"] == "t.b.d.":
        value = "n.t.b."
    else:
        value = r["bound"]["value"].split(".")[-1]
    unit = ""
    if r["bound"]["unit"]:
        unit = r["bound"]["unit"]

    if r["auxiliary"] == "won't":
        return "{} {} nimmer {} {} {} {}".format(s, aux[0], comparison, value, unit, aux[1])
    else:
        return "{} {} {} {} {} {}".format(aux[0], s, comparison, value, unit, aux[1])


srule_auxmap = {
    "must be": "dient te zijn",
    "shall be": "zal zijn",
    "should be": "behoort te zijn",
    "could be": "zou kunnen zijn",
    "won't": "zal zijn",
    "is": "is",
}


def subclause_rule_line(r: Dict[str, str]):
    s = r["subject"].split(".")[-1]
    aux = srule_auxmap.get(r["auxiliary"])
    comparison = cmap.get(r["comparison"])

    if is_number(v=r["bound"]["value"]):
        value = r["bound"]["value"]
    elif r["bound"]["value"] == "t.b.d.":
        value = "n.t.b."
    else:
        value = r["bound"]["value"].split(".")[-1]
    unit = ""
    if r["bound"]["unit"]:
        unit = r["bound"]["unit"]

    if r["auxiliary"] == "won't":
        return "{} nimmer {} {} {} {}".format(s, comparison, value, unit, aux)
    else:
        return "{} {} {} {} {}".format(s, comparison, value, unit, aux)


@hookimpl
def function_spec_node(node: Node, graph: Graph) -> LineGen:
    if node.annotations.esl_info["sub_kind"] == "transformation":
        comp = node.annotations.esl_info["body"]["active"].split(".")[-1]
        aux = node.annotations.esl_info["body"]["auxiliary"]
        verb = node.annotations.esl_info["body"]["verb"]
        in_flows = pm.hook.linguistic_enumeration(
            items=sorted(
                list(
                    set(
                        [
                            graph[flow].annotations.esl_info["bundle_root_name"]
                            if graph[flow].annotations.esl_info["bundle_root_name"]
                            else flow.split(".")[-1]
                            for flow in node.annotations.esl_info["body"]["input_variables"]
                        ]
                    )
                )
            )
        )
        prep = node.annotations.esl_info["body"]["preposition"]
        out_flows = pm.hook.linguistic_enumeration(
            items=sorted(
                list(
                    set(
                        [
                            graph[flow].annotations.esl_info["bundle_root_name"]
                            if graph[flow].annotations.esl_info["bundle_root_name"]
                            else flow.split(".")[-1]
                            for flow in node.annotations.esl_info["body"]["output_variables"]
                        ]
                    )
                )
            )
        )
        if aux == "must":
            line = "{} dient {} {} {} {}".format(
                comp, in_flows, prep, out_flows, make_predicate(verb)
            )
        elif aux == "should":
            line = "{} behoort {} {} {} {}".format(
                comp, in_flows, prep, out_flows, make_predicate(verb)
            )
        elif aux == "could":
            line = "{} zou {} {} {} kunnen {}".format(comp, in_flows, prep, out_flows, verb)
        elif aux == "won't":
            line = "{} zal nimmer {} {} {} {}".format(comp, in_flows, prep, out_flows, verb)
        elif aux == "shall":
            line = "{} zal {} {} {} {}".format(comp, in_flows, prep, out_flows, verb)
        elif aux == "is":
            line = "{} vervult het {} van {} {} {}".format(comp, verb, in_flows, prep, out_flows)
        else:
            line = "{} {} {} {} {} {}".format(comp, aux, verb, in_flows, prep, out_flows)

        # Bundle root varaible dict
        brvdict = defaultdict(list)
        for flow in (
            node.annotations.esl_info["body"]["input_variables"]
            + node.annotations.esl_info["body"]["output_variables"]
        ):
            r = graph[flow].annotations.esl_info["bundle_root_name"]
            if r:
                brvdict[r].append(flow)

    elif node.annotations.esl_info["sub_kind"] == "goal":
        active = node.annotations.esl_info["body"]["active"].split(".")[-1]
        aux = node.annotations.esl_info["body"]["auxiliary"]
        verb = node.annotations.esl_info["body"]["verb"]
        flows = pm.hook.linguistic_enumeration(
            items=sorted(
                list(
                    set(
                        [
                            graph[flow].annotations.esl_info["bundle_root_name"]
                            if graph[flow].annotations.esl_info["bundle_root_name"]
                            else flow.split(".")[-1]
                            for flow in node.annotations.esl_info["body"]["variables"]
                        ]
                    )
                )
            )
        )
        prep = node.annotations.esl_info["body"]["preposition"]
        passive = node.annotations.esl_info["body"]["passive"].split(".")[-1]

        if aux == "must":
            line = "{} dient {} {} {} {}".format(active, flows, prep, passive, make_predicate(verb))
        elif aux == "should":
            line = "{} behoort {} {} {} {}".format(
                active, flows, prep, passive, make_predicate(verb)
            )
        elif aux == "could":
            line = "{} zou {} {} {} kunnen {}".format(active, flows, prep, passive, verb)
        elif aux == "won't":
            line = "{} zal nimmer {} {} {} {}".format(active, flows, prep, passive, verb)
        elif aux == "does":
            line = "{} vervult het {} van {} {} {}".format(active, verb, flows, prep, passive)
        elif aux == "shall":
            line = "{} zal {} {} {} {}".format(active, flows, prep, passive, verb)
        else:
            line = "{} {} {} {} {} {}".format(
                active, aux, make_predicate(verb), flows, prep, passive
            )

        # Bundle root varaible dict
        brvdict = defaultdict(list)
        for flow in node.annotations.esl_info["body"]["variables"]:
            r = graph[flow].annotations.esl_info["bundle_root_name"]
            if r:
                brvdict[r].append(flow)

    if node.annotations.esl_info["body"]["subclauses"]:
        yield lns.cap(line) + ", waarbij:\n"
        yield from lns.unordered(
            [
                "\n".join(subclause_line(s=sc))
                for sc in node.annotations.esl_info["body"]["subclauses"]
            ]
        )
    else:
        yield lns.snt(line)

    if brvdict:
        yield from lns.bundle_clarification(brvdict=brvdict)


@hookimpl
def design_spec_node(node: Node, graph: Graph) -> LineGen:
    line = " of ".join([design_rule_line(r) for r in node.annotations.esl_info["body"]])
    if node.annotations.esl_info.get("sub_clauses"):
        yield lns.cap(line) + ", waarbij:\n"
        yield from lns.unordered(
            ["\n".join(subclause_line(s=sc)) for sc in node.annotations.esl_info["sub_clauses"]]
        )
    else:
        yield lns.snt(line)

    bvars = [
        v
        for v in lns.get_design_rule_line_vars(rules=node.annotations.esl_info["body"], g=graph)
        if v.annotations.esl_info["bundle_root_name"]
    ]

    if bvars:
        yield from lns.var_clarification(bvars=bvars)


@hookimpl
def behavior_spec_node(node: Node, graph: Graph) -> LineGen:
    clauses = []
    for case in node.annotations.esl_info["cases"]:
        yield lns.cap("situatie {}:\n".format(lns.emph(case["name"])))
        yield ""
        yield "als:\n"
        yield from lns.unordered(["\n".join(subclause_line(s=sc)) for sc in case["when_clauses"]])
        yield "dan:\n"
        yield from lns.unordered(["\n".join(then_clause_line(s=sc)) for sc in case["then_clauses"]])
        clauses += case["when_clauses"] + case["then_clauses"]

    default = node.annotations.esl_info.get("default")
    if default:
        yield lns.cap("situatie {}:\n".format(lns.emph("default")))
        yield ""
        yield "als geen andere situatie van toepassing is, dan:\n"
        yield from lns.unordered(["\n".join(then_clause_line(s=sc)) for sc in default])
        clauses += default

    bvars = [
        v
        for s in clauses
        for v in lns.get_design_rule_line_vars(rules=s["body"], g=graph)
        if v.annotations.esl_info["bundle_root_name"]
    ]

    if bvars:
        yield from lns.var_clarification(bvars=bvars)


@hookimpl
def need_node(node: Node, graph: Graph) -> LineGen:
    subject = node.annotations.esl_info["subject"]
    line = "{} {}".format(
        node.annotations.esl_info["subject"].split(".")[-1],
        node.annotations.esl_info["text"],
    )
    yield lns.snt(line)

    if graph[subject].kind == "variable":
        if graph[subject].annotations.esl_info["bundle_root_name"]:
            yield from lns.var_clarification(bvars=[graph[subject]])



"""Excel sheet generation."""

from typing import List, Tuple

from openpyxl.workbook import Workbook
from openpyxl.worksheet.worksheet import Worksheet
from raesl.doc import utils as doc_utils
from raesl.doc.lines import node_path, var_path
from raesl.doc.sections import get_var_table_row_elements
from raesl.excel import text, utils
from raesl.excel.defaults import OPTIONS
from ragraph.graph import Graph
from ragraph.node import Node


def add_components_sheet(wb: Workbook, components: List[Node]) -> Tuple[Worksheet, List[Node]]:
    """Add a components overview sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the components sheet to.
        components: List of component nodes.

    Returns:
        Components worksheet instance.
    """
    ws = wb.create_sheet("Components")

    # Handle headers.
    tags = utils.get_all_tags(components)
    headers = [
        "instance path",
        "component definition",
        "parent component",
        "comments",
    ] + tags
    ws.append(headers)

    # Handle content.
    def write_component_row(ws: Worksheet, tags: List[str], component: Node):
        """Write a component sheet row."""
        info = component.annotations.esl_info
        row = [
            node_path(component.name, italic=False, arrows=False, skip="world"),
            info["definition_name"],
            node_path(component.parent.name, italic=False, arrows=False, skip="world")
            if component.parent
            else None,
            utils.format_multiline(info["comments"]),
        ]
        for tag in tags:
            comments = info["tagged_comments"].get(tag, [])
            row.append(utils.format_multiline(comments))
        ws.append(row)

    for component in components:
        write_component_row(ws, tags, component)

    # Handle styling.
    utils.apply_styling(ws, headers, defaults=OPTIONS)
    utils.make_table(
        ws,
        name="components",
        min_row=1,
        max_row=len(components) + 1,
        min_col=1,
        max_col=len(headers),
    )
    return ws, components


def add_goals_sheet(
    wb: Workbook, graph: Graph, components: List[Node]
) -> Tuple[Worksheet, List[Node]]:
    """Add a goal requirements sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the goals sheet to.
        graph: Graph to fetch goals from.
        components: Component nodes to fetch goals for.

    Returns:
        Goal requirements worksheet instance.
    """
    ws = wb.create_sheet("Goals")

    # Get requirements.
    requirements = utils.dedupe(
        [
            r
            for c in components
            for r in doc_utils.get_component_goals(c, graph, constraint=False, inherited=False)
        ]
        + [
            r
            for c in components
            for r in doc_utils.get_component_goals(c, graph, constraint=False, inherited=True)
        ]
        + [
            r
            for c in components
            for r in doc_utils.get_component_goals(c, graph, constraint=True, inherited=False)
        ]
        + [
            r
            for c in components
            for r in doc_utils.get_component_goals(c, graph, constraint=True, inherited=True)
        ]
    )

    # Handle headers.
    default_headers = [
        "instance path",
        "component definition",
        "form",
        "source component",
        "auxiliary",
        "verb",
        "variables",
        "preposition",
        "target component",
        "subclauses",
        "comments",
    ]
    tags = utils.get_all_tags(requirements)
    headers = default_headers + tags
    ws.append(headers)

    # Handle content.
    for requirement in requirements:
        info = requirement.annotations.esl_info
        body = info["body"]

        default_content = [
            node_path(requirement.name, italic=False, arrows=False, skip="world"),
            utils.parent_def(graph, requirement),
            info["form"],
            node_path(body["active"], italic=False, arrows=False, skip="world"),
            body["auxiliary"],
            body["verb"],
            ", ".join(
                var_path(
                    graph[var],
                    italic=False,
                    arrows=False,
                    skip="world",
                )
                for var in body["variables"]
            ),
            body["preposition"],
            node_path(body["passive"], italic=False, arrows=False, skip="world"),
            text.subclauses_text(requirement, graph, skip="world", spaces=0),
            utils.format_multiline(info["comments"]),
        ]
        tagged_content = [
            utils.format_multiline(info["tagged_comments"].get(tag, [])) for tag in tags
        ]
        ws.append(default_content + tagged_content)

    # Handle styling.
    utils.make_table(
        ws,
        "goals",
        min_row=1,
        max_row=1 + len(requirements),
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS)
    return ws, requirements


def add_transformations_sheet(
    wb: Workbook, graph: Graph, components: List[Node]
) -> Tuple[Worksheet, List[Node]]:
    """Add a transformation requirements sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the transformations sheet to.
        graph: Graph to fetch transformation nodes from.
        components: Component nodes to fetch transformations for.

    Returns:
        Transformation requirements worksheet instance.
    """
    ws = wb.create_sheet("Transformations")

    # Select requirements.
    requirements = utils.dedupe(
        [
            r
            for c in components
            for r in doc_utils.get_component_transformations(c, graph, constraint=False)
        ]
        + [
            r
            for c in components
            for r in doc_utils.get_component_transformations(c, graph, constraint=True)
        ]
    )

    # Handle headers.
    default_headers = [
        "instance path",
        "component definition",
        "form",
        "auxiliary",
        "verb",
        "input_variables",
        "preposition",
        "output_variables",
        "subclauses",
        "comments",
    ]
    tags = utils.get_all_tags(requirements)
    headers = default_headers + tags
    ws.append(headers)

    for requirement in requirements:
        info = requirement.annotations.esl_info
        body = info["body"]

        default_content = [
            node_path(requirement.name, italic=False, arrows=False, skip="world"),
            utils.parent_def(graph, requirement),
            info["form"],
            body["auxiliary"],
            body["verb"],
            ", ".join(
                var_path(graph[var], italic=False, arrows=False, skip="world")
                for var in body["input_variables"]
            ),
            body["preposition"],
            ", ".join(
                var_path(graph[var], italic=False, arrows=False, skip="world")
                for var in body["output_variables"]
            ),
            text.subclauses_text(requirement, graph, skip="world", spaces=0),
            utils.format_multiline(info["comments"]),
        ]
        tagged_content = [
            utils.format_multiline(info["tagged_comments"].get(tag, [])) for tag in tags
        ]
        ws.append(default_content + tagged_content)

    # Handle styling.
    utils.make_table(
        ws,
        "tranformations",
        min_row=1,
        max_row=1 + len(requirements),
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS)

    return ws, requirements


def add_designs_sheet(
    wb: Workbook, graph: Graph, components: List[Node]
) -> Tuple[Worksheet, List[Node]]:
    """Add a design requirements sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the design requirements sheet to.
        graph: Graph to fetch designrule nodes from.
        components: Component nodes to fetch design requirements for.

    Returns:
        Design requirements worksheet instance.
    """
    ws = wb.create_sheet("Design requirements")

    # Select requirements.
    requirements = utils.dedupe(
        [r for c in components for r in doc_utils.get_component_designs(c, graph, constraint=False)]
        + [
            r
            for c in components
            for r in doc_utils.get_component_designs(c, graph, constraint=True)
        ]
    )

    # Handle headers.
    default_headers = [
        "instance path",
        "component definition",
        "form",
        "subject",
        "auxiliary",
        "comparison",
        "bound",
        "subclauses",
        "comments",
    ]
    tags = utils.get_all_tags(requirements)
    headers = default_headers + tags
    ws.append(headers)
    rows = 1

    # Handle content.
    for requirement in requirements:
        info = requirement.annotations.esl_info
        for body in info["body"]:
            default_content = [
                node_path(requirement.name, italic=False, arrows=False, skip="world"),
                utils.parent_def(graph, requirement),
                info["form"],
                var_path(graph[body["subject"]], italic=False, arrows=False, skip="world"),
                "{}{}".format("EITHER " if len(info["body"]) > 1 else "", body["auxiliary"]),
                body["comparison"],
                "{} {}".format(body["bound"]["value"], body["bound"]["unit"]),
                text.subclauses_text(requirement, graph, skip="world", spaces=0),
                utils.format_multiline(info["comments"]),
            ]
            tagged_content = [
                utils.format_multiline(info["tagged_comments"].get(tag, [])) for tag in tags
            ]
            ws.append(default_content + tagged_content)
            rows += 1

    # Handle styling.
    utils.make_table(
        ws,
        "designrequirements",
        min_row=1,
        max_row=rows,
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS)
    return ws, requirements


def add_behaviors_sheet(
    wb: Workbook, graph: Graph, components: List[Node]
) -> Tuple[Worksheet, List[Node]]:
    """Add a behavior requirements sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the behavior requirements sheet to.
        graph: Graph to fetch behavior nodes from.
        components: Component nodes to fetch behavior requirements for.

    Returns:
        Behavior requirements worksheet instance.
    """
    ws = wb.create_sheet("Behavior requirements")

    # Select requirements.
    requirements = utils.dedupe(
        [
            r
            for c in components
            for r in doc_utils.get_component_behaviors(c, graph, constraint=False)
        ]
        + [
            r
            for c in components
            for r in doc_utils.get_component_behaviors(c, graph, constraint=True)
        ]
    )

    # Handle headers.
    default_headers = [
        "instance path",
        "component definition",
        "form",
        "case",
        "when",
        "then",
        "comments",
    ]
    tags = utils.get_all_tags(requirements)
    headers = default_headers + tags
    ws.append(headers)
    rows = 1

    # Handle content.
    for requirement in requirements:
        info = requirement.annotations.esl_info
        tagged_content = [
            utils.format_multiline(info["tagged_comments"].get(tag, [])) for tag in tags
        ]
        if info["default"]:
            default_content = [
                node_path(requirement.name, italic=False, arrows=False, skip="world"),
                utils.parent_def(graph, requirement),
                info["form"],
                "default",
                "no other case applies",
                "\n".join(
                    text.designclause_text(clause["body"], graph, label=clause["name"])
                    for clause in info["default"]
                ),
                utils.format_multiline(info["comments"]),
            ]
            ws.append(default_content + tagged_content)
            rows += 1
        for case in info["cases"]:
            default_content = [
                node_path(requirement.name, italic=False, arrows=False, skip="world"),
                utils.parent_def(graph, requirement),
                info["form"],
                case["name"],
                "\n".join(
                    text.designclause_text(clause["body"], graph, label=clause["name"])
                    for clause in case["when_clauses"]
                ),
                "\n".join(
                    text.designclause_text(clause["body"], graph, label=clause["name"])
                    for clause in case["then_clauses"]
                ),
                utils.format_multiline(info["comments"]),
            ]
            ws.append(default_content + tagged_content)
            rows += 1

    # Handle styling.
    utils.make_table(
        ws,
        "behaviors",
        min_row=1,
        max_row=rows,
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS)
    return ws, requirements


def add_needs_sheet(
    wb: Workbook, graph: Graph, components: List[Node]
) -> Tuple[Worksheet, List[Node]]:
    """Add a needs sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the needs sheet to.
        graph: Graph to fetch need nodes from.
        components: Component nodes to fetch needs for.

    Returns:
        Needs worksheet instance.
    """
    ws = wb.create_sheet("Needs")

    # Select requirements.
    requirements = utils.dedupe(
        r for c in components for r in doc_utils.get_component_needs(c, graph)
    )

    # Handle headers.
    default_headers = [
        "instance path",
        "component definition",
        "subject",
        "text",
    ]
    tags = utils.get_all_tags(requirements)
    headers = default_headers + tags
    ws.append(headers)

    # Handle content.
    for requirement in requirements:
        info = requirement.annotations.esl_info
        default_content = [
            requirement.name.split(".")[-1],
            utils.parent_def(graph, requirement),
            var_path(graph[info["subject"]], italic=False, arrows=False, skip="world"),
            info["text"],
            utils.format_multiline(info["comments"]),
        ]
        tagged_content = [
            utils.format_multiline(info["tagged_comments"].get(tag, [])) for tag in tags
        ]
        ws.append(default_content + tagged_content)

    # Handle styling.
    utils.make_table(
        ws,
        "needs",
        min_row=1,
        max_row=1 + len(requirements),
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS)
    return ws, requirements


def add_variable_sheet(
    wb: Workbook, graph: Graph, components: List[Node]
) -> Tuple[Worksheet, List[Node]]:
    """Add a variable sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the needs sheet to.
        graph: Graph to fetch need nodes from.
        components: Component nodes to fetch needs for.

    Returns:
        Needs worksheet instance.
    """
    ws = wb.create_sheet("Variables")

    # Variables
    vrs = [n for n in graph.nodes if n.kind == "variable"]
    vrs = sorted(vrs, key=lambda x: x.name)

    # Handle headers.
    default_headers = ["instance path", "type", "domain", "units", "clarifaction"]
    tags = utils.get_all_tags(vrs)
    headers = default_headers + tags
    ws.append(headers)

    # Handle content.
    for var in vrs:
        info = var.annotations.esl_info
        default_content = list(get_var_table_row_elements(graph, var))
        tagged_content = [
            utils.format_multiline(info["tagged_comments"].get(tag, [])) for tag in tags
        ]
        ws.append(default_content + tagged_content)

    # Handle styling.
    utils.make_table(
        ws,
        "variables",
        min_row=1,
        max_row=1 + len(vrs),
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS)
    return ws, vrs


def add_overview_sheet(
    wb: Workbook,
    graph: Graph,
    components: List[Node],
    goals: List[Node],
    transformations: List[Node],
    designs: List[Node],
    behaviors: List[Node],
    needs: List[Node],
) -> Worksheet:
    """Add an overview sheet to an Excel workbook.

    Arguments:
        wb: Excel workbook to add the needs sheet to.
        components: Component nodes.
        goals: Goal requirement nodes.
        transformations: Transformation requirement nodes.
        designs: Design requirement nodes.
        behaviors: Behavior requirement nodes.
        needs: Need nodes.

    Returns:
        Overview worksheet instance.
    """
    ws = wb.create_sheet("Overview", index=0)

    # Select requirements.
    requirements = goals + transformations + designs + behaviors + needs

    # Handle headers.
    default_headers = [
        "instance name",
        "specification text",
        "component path",
        "component definition",
        "kind",
        "form",
        "comments",
    ]
    tags = utils.get_all_tags(components + requirements)
    headers = default_headers + tags
    ws.append(headers)

    # Handle content.
    for requirement in requirements:
        info = requirement.annotations.esl_info
        skip = utils.parent_component(requirement, skip=None)
        default_content = [
            requirement.name.split(".")[-1],
            text.requirement_text(requirement, graph, skip=skip),
            utils.parent_component(requirement, skip="world"),
            utils.parent_def(graph, requirement),
            utils.requirement_kind(requirement),
            None if requirement.kind == "need" else info["form"],
            utils.format_multiline(info["comments"]),
        ]
        tagged_content = [
            utils.format_multiline(info["tagged_comments"].get(tag, [])) for tag in tags
        ]
        ws.append(default_content + tagged_content)

    utils.make_table(
        ws,
        "overview",
        min_row=1,
        max_row=1 + len(requirements),
        min_col=1,
        max_col=len(headers),
    )
    utils.apply_styling(ws, headers, defaults=OPTIONS)

    return ws



"""ESL to Excel Command Line Interface."""

import sys
from typing import Iterable, Tuple

import click
import raesl.excel
import raesl.excel.defaults
from raesl import logger


@click.command("excel")
@click.argument("paths", nargs=-1, type=click.Path(exists=True, file_okay=True, dir_okay=True))
@click.option(
    "--output",
    "-o",
    default=raesl.excel.defaults.OUTPUT,
    type=click.Path(file_okay=True, dir_okay=False, writable=True),
    help="Output file to write to.",
    show_default=True,
)
@click.option(
    "--scope",
    "-s",
    default=[("world", -1)],
    type=(str, int),
    multiple=True,
    help=(
        "Scopes in the component hierarchy. "
        "Expects component instance paths and the relative depth you want to include. "
        "A depth of -1 includes everything below that component."
    ),
    show_default=True,
)
def excel(paths: Iterable[str], output: str, scope: Iterable[Tuple[str, int]]):
    """Convert ESL files and/or directories to an Excel workbook."""
    logger.info("This is the Ratio ESL Excel command line utility.")
    logger.info(f"Populating '{output}'...")
    try:
        raesl.excel.convert(
            *paths,
            output=output,
            scopes={s[0]: s[1] if s[1] >= 0 else None for s in scope},
        )
        logger.info("Excel generation done!")
    except Exception as e:
        logger.error(str(e))
        sys.exit(1)


if __name__ == "__main__":
    excel()



"""Variables and components in a component definition."""
from typing import TYPE_CHECKING, Any, Dict, List

from raesl.compile import diagnostics
from raesl.compile.ast import components, nodes, types

if TYPE_CHECKING:
    from raesl.compile.ast.specification import Specification
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.compdef_builder import CompDefChildBuilders, Counter


class CompDefVarParamBuilder:
    """Collect and check variables of a component definition.

    Arguments:
        comp_child_builders: Storage of child builders for a component definition.
        varparam_counter: Object for handing out unique numbers to elementary var/param
            nodes.

    Attributes:
        variables: Variables of the component.
        parameters: Parameters of the component.
    """

    def __init__(self, comp_child_builders: "CompDefChildBuilders", varparam_counter: "Counter"):
        self.diag_store: diagnostics.DiagnosticStore = comp_child_builders.diag_store
        self.varparam_counter = varparam_counter
        self.comp_child_builders = comp_child_builders

        self.variables: List[components.VarParam] = []
        self.parameters: List[components.VarParam] = []

    def add_variables(self, new_vars: List[components.VarParam]):
        """Add variables of the component definition to the collection."""
        self.variables.extend(new_vars)

    def add_parameters(self, new_params: List[components.VarParam]):
        """Add parameters of the component definition to the collection."""
        self.parameters.extend(new_params)

    def finish_comp(self, comp_def: components.ComponentDefinition, spec: "Specification"):
        """Check the collected variables and parameters, report errors, and add the
        instances to the given component.

        Arguments:
            comp_def: Component definition to extend with the found variables and
                parameters.
            spec: Specification being constructed. Source for types, verbs and relation
                definitions processed previously.
        """
        varsparams_by_name: Dict[str, List[Any]]
        varsparams_by_name = self.comp_child_builders.elements_by_label
        for var in self.variables:
            name = var.name_tok.tok_text
            varsparams_by_name[name].append(var)
        for param in self.parameters:
            name = param.name_tok.tok_text
            varsparams_by_name[name].append(param)

        for name, varsparams in varsparams_by_name.items():
            # Verify properties of the variable or parameter.
            varparam = varsparams[0]
            typename = varparam.type_tok.tok_text
            vartypedef = spec.types.get(typename)
            if vartypedef is None:
                loc = varparam.name_tok.get_location()
                self.diag_store.add(diagnostics.E203("type", name=typename, location=loc))
            else:
                # Build node tree for the variable or parameter.
                varparam.node = _make_varnodes(
                    varparam.name_tok, vartypedef.type, self.varparam_counter
                )

                # And store it at the right spot in the component definition.
                if varparam.is_variable:
                    comp_def.variables.append(varparam)
                else:
                    comp_def.parameters.append(varparam)


def _make_varnodes(name_tok: "Token", the_type: types.BaseType, varparam_counter: "Counter"):
    """Make a VarNode tree that matches the shape of the type."""
    if isinstance(the_type, types.ElementaryType):
        return nodes.ElementaryVarNode(name_tok, the_type, varparam_counter)
    else:
        assert isinstance(the_type, types.Compound)
        childs = [_make_varnodes(fld.name, fld.type, varparam_counter) for fld in the_type.fields]
        return nodes.CompoundVarNode(name_tok, the_type, childs)



"""State machine classes to describe an allowed sequences of tokens.

A state machine is a DFA (deterministic finite automaton, always at most one
edge that matches). An edge is associated with a matched token, locations are
decision points between tokens.

An edge may tag occurrences of tokens to simplify extraction of relevant
information for future compiler phases. A location may record matching of
a valid sequence of tokens by accepting.
"""
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Tuple

from raesl.compile.scanner import get_token_priority

if TYPE_CHECKING:
    from raesl.compile.scanner import Lexer, Token


class Location:
    """Location in a state machine.

    Arguments:
        accept: Name of the rule that could be accepted at this location.
            If None, no such rule exists.
        name: Name of the location, mostly for identifying purposes.

    Attributes:
        out_edges: Outgoing edges, initially empty.
    """

    def __init__(self, name: str, accept: Optional[str] = None):
        self.accept = accept
        self.name = name
        self.out_edges: List[Edge] = []


class Edge:
    """Edge to a next location.

    Arguments:
        dest: Destination of the edge.
        tok_type: The value of the 'tok_type' attribute of a token that can trigger
            this transition.
        tag_name: If not None, the name to use for recording the transition in the
            state machine.
    """

    def __init__(self, dest: Location, tok_type: str, tag_name: Optional[str] = None):
        self.tok_type = tok_type
        self.tag_name = tag_name
        self.dest = dest
        assert isinstance(dest, Location)


class StateMachine:
    """State machine containing locations and edges.

    Note that it only stores the initial location, all other locations and edges
    are reachable from it.

    Arguments:
        name: Name of the state machine, also the name of the matched sequence.

    Attributes:
        initial_loc: Initial location of the state machine. Set after construction.
    """

    def __init__(self, name: str):
        self.name = name
        self.initial_loc: Optional[Location] = None

    def sort_edges(self):
        """Sort edges of the state machine to get specific tokens checked first."""
        done_locs = set([self.initial_loc])
        found_locs = [self.initial_loc]
        while found_locs:
            loc = found_locs.pop()
            loc.out_edges.sort(key=lambda edge: get_token_priority(edge.tok_type))
            for edge in loc.out_edges:
                if edge.dest not in done_locs:
                    done_locs.add(edge.dest)
                    found_locs.append(edge.dest)

    def dump(self, fname: Optional[str] = None):
        """Dump the state machine to a file in Graphviz format.

        Arguments:
            fname: If not None, name of the file to write, else a filename is
                constructed from the name of the state machine.
        """
        processed_locs: Set[Location] = set()
        notdone = [self.initial_loc]
        lines = ["digraph G {"]
        while notdone:
            loc = notdone.pop()
            if loc is None or loc in processed_locs:
                continue

            processed_locs.add(loc)
            if loc.accept:
                lines.append(f'    {loc.name} [shape="box"]')
            else:
                lines.append(f"    {loc.name}")

            for edge in loc.out_edges:
                lines.append(f'    {loc.name} -> {edge.dest.name} [label="{edge.tok_type}"]')
                notdone.append(edge.dest)

        lines.append("}")

        if fname is None:
            fname = self.name + ".dot"
        with open(fname, "w") as handle:
            for line in lines:
                handle.write(line)
                handle.write("\n")

    def match(self, lexer: "Lexer") -> Optional["MatchResult"]:
        """Try to match the machine against tokens from the scanner.

        Arguments:
            lexer: Token stream to match. Instance is useless afterwards, make a copy
                beforehand if you need it again.

        Returns:
            Result of the matching process.

        Note:
            This routine is currently only used for testing.
        """
        assert self.initial_loc
        current_loc: Location = self.initial_loc
        tags: Dict[str, List["Token"]] = {}

        accepted_name = None
        accepted_tags: Dict[str, List["Token"]] = {}
        accepted_lexer: Optional["Lexer"] = None

        while True:
            # Keep stepping until no progress is possible any more.
            match = self.single_step(lexer, current_loc, tags)
            if not match:
                break

            current_loc = match[0]

            # Update acceptance if necessary.
            if current_loc.accept is not None:
                accepted_name = current_loc.accept
                accepted_tags = dict((k, v.copy()) for k, v in tags.items())
                accepted_lexer = lexer.copy()

        # Done, return the result, either just the lexer at the failed state
        # for its position information, or the last accept.
        if accepted_name is None:
            return MatchResult(None, {}, lexer)

        assert accepted_lexer
        return MatchResult(accepted_name, accepted_tags, accepted_lexer)

    def single_step(
        self, lexer: "Lexer", current_loc: Location, tags: Dict[str, List["Token"]]
    ) -> Optional[Tuple["Location", "Token"]]:
        """Try to perform a single step in the state machine.

        Arguments:
            lexer: Token stream to match. Instance is modified in-place if a transition
                is taken.
            current_loc: Location to use for finding edges to try.
            tags: Collected tags so far, may be updated in-place if transition was
                performed.

        Returns:
            New location and the matching token if a transition could be performed,
            else None.
        """
        if not current_loc.out_edges:
            return None

        for edge in current_loc.out_edges:
            token = lexer.find(edge.tok_type)
            if token is None:
                continue

            # Match found. Due to being a DFA and edges being sorted on priority,
            # this is also the one and only match that we should find for this machine.
            #
            # Update tags
            if edge.tag_name is not None:
                edge_tag = tags.get(edge.tag_name)
                if edge_tag is None:
                    tags[edge.tag_name] = [token]
                else:
                    edge_tag.append(token)

            # Pass target location back to the caller.
            return edge.dest, token

        return None


class MatchResult:
    """Result data of a matching process in StateMachine.match().

    Arguments:
        accepted_name: Acceptance name from the last visited accepting location.
        accepted_tags: Collected tag data at the point of accepting.
        lexer: Lexer at the time of the last fail or at the time of the last accept
    """

    def __init__(
        self,
        accepted_name: Optional[str],
        accepted_tags: Dict[str, List["Token"]],
        lexer: "Lexer",
    ):
        self.accepted_name = accepted_name
        self.accepted_tags = accepted_tags
        self.lexer = lexer



"""Excel export utility methods."""
from typing import Any, Dict, Iterable, List, Optional, Set

from openpyxl.utils.cell import get_column_letter
from openpyxl.worksheet.cell_range import CellRange
from openpyxl.worksheet.table import Table, TableStyleInfo
from openpyxl.worksheet.worksheet import Worksheet
from raesl.doc import lines
from raesl.excel.defaults import MONO, WRAP
from ragraph.graph import Graph
from ragraph.node import Node


def get_all_tags(nodes: List[Node]) -> List[str]:
    """Get all tagged comment keys from a list of nodes."""
    tags: List[str] = []
    seen: Set[str] = set()
    for node in nodes:
        node_tags = [
            tag for tag in node.annotations.esl_info["tagged_comments"].keys() if tag not in seen
        ]
        tags.extend(node_tags)
        seen.update(node_tags)
    return tags


def make_table(
    ws: Worksheet,
    name: str = "Table",
    min_row: int = 1,
    max_row: int = 1,
    min_col: int = 1,
    max_col: int = 1,
    style: TableStyleInfo = TableStyleInfo(name="TableStyleMedium9", showRowStripes=True),
) -> Table:
    """Make a table of a cell range in a worksheet."""
    ref = CellRange(min_row=min_row, max_row=max_row, min_col=min_col, max_col=max_col)
    table = Table(name=name, displayName=name, ref=ref.coord)
    table.tableStyleInfo = style
    ws.add_table(table)
    return table


def apply_styling(ws: Worksheet, headers: List[str], defaults: Dict[str, Any] = dict()):
    """Apply styling to columns given some default option dictionary."""
    for i, header in enumerate(headers):
        idx = get_column_letter(i + 1)
        for cell in ws[idx]:
            styles = defaults.get(header, dict()).get("styles", dict())
            cell.font = styles.get("font", MONO)
            cell.alignment = styles.get("alignment", WRAP)
        width = defaults.get(header, dict()).get("width", 60)
        if width is not None:
            ws.column_dimensions[idx].width = width


def parent_component(requirement: Node, skip: Optional[str] = "world") -> str:
    """Get parent component name of a requirement."""
    path = ".".join(requirement.name.split(".")[:-1])
    return lines.node_path(path, arrows=False, skip=skip) if skip else path


def parent_def(graph: Graph, requirement: Node) -> str:
    """Get the parent (component) definition of a requirement."""
    parent_comp = ".".join(requirement.name.split(".")[:-1])
    parent_def = graph[parent_comp].annotations.esl_info["definition_name"]
    return parent_def


def format_multiline(comments: List[str]) -> str:
    """Format multiline (list) text."""
    return "\n".join(c.rstrip("\\") for c in comments)


def requirement_kind(requirement: Node) -> str:
    """Get requirement kind."""
    info = requirement.annotations.esl_info
    if requirement.kind == "function_spec":
        return info["sub_kind"]
    elif requirement.kind == "design_spec":
        return "design"
    elif requirement.kind == "behavior_spec":
        return "behavior"
    else:
        return requirement.kind


def dedupe(iterable: Iterable) -> List[Any]:
    """Deduplicate any iterable into a list where the first occurrence is preserved."""
    seen: Set[Any] = set()
    unique: List[Any] = list()
    for item in iterable:
        if item in seen:
            continue
        seen.add(item)
        unique.append(item)
    return unique



"""Collect and process relation definition that are found by the parser."""
from collections import defaultdict
from typing import TYPE_CHECKING, Dict, List, Optional

from raesl.compile import diagnostics
from raesl.compile.ast import relations

if TYPE_CHECKING:
    from raesl.compile.ast.specification import Specification
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.ast_builder import AstBuilder

INPUT = relations.INPUT
OUTPUT = relations.OUTPUT
INPOUT = relations.INPOUT


class RelationDefBuilder:
    """Builder to construct relation definitions.

    Arguments:
        ast_builder: AST builder instance.

    Attributes:
        diag_store: Storage for found diagnostics.
        rel_defs: Created relation definitions while collecting data from parsing.
        current_reldef: Reference to the entry in 'rel_defs' that is being filled.
        last_occurrences: Map of input/output directions to token of last occurrence
            of that direction in the current definition.
        current_direction: Parameter direction to attach to a new parameter.
    """

    def __init__(self, ast_builder: "AstBuilder"):
        self.diag_store = ast_builder.diag_store

        self.rel_defs: Optional[List[relations.RelationDefinition]] = []
        self.current_reldef: Optional[relations.RelationDefinition] = None
        self.last_occurrences: Dict[str, "Token"] = {}
        self.current_direction: Optional[str] = None

        ast_builder.register_new_section(self)

    def notify_new_section(self, _new_top_section: bool):
        """Parser found a new section, drop all 'in-progress' relation definition
        construction.
        """
        self.current_reldef = None
        self.last_occurrences = {}
        self.current_direction = None

    def add_reldef(self, name: "Token"):
        """Add a new relation definition. Parameters will follow.

        Arguments:
            Name of the relation definition.
        """
        assert self.rel_defs is not None

        reldef = relations.RelationDefinition(name)
        self.rel_defs.append(reldef)
        self.current_reldef = reldef
        self.last_occurrences = {}
        self.current_direction = None

    def reldef_param_header(self, header_tok: "Token", direction: str):
        """New parameter subsection with a direction. Set the direction for the
        parameters that will follow.

        Arguments:
            header_tok: Token of the direction, for deriving position information if
                needed.
            direction: Direction of the next parameters of the relation definition.
        """
        assert self.current_reldef is not None

        last_occurrence = self.last_occurrences.get(direction)
        if last_occurrence is not None:
            direction_text = {
                INPUT: "require",
                OUTPUT: "returning",
                INPOUT: "relating",
            }[direction]
            locs = [header_tok.get_location(), last_occurrence.get_location()]
            self.diag_store.add(
                diagnostics.E200(direction_text, "parameter section", location=locs[0], dupes=locs)
            )
            # Continue anyway

        self.last_occurrences[direction] = header_tok
        self.current_direction = direction

    def reldef_add_param(self, name: "Token", type_name: "Token", multi: bool):
        """Add a parameter to the current relation definition."""
        assert self.current_direction is not None
        assert self.current_reldef is not None

        rel_param = relations.RelationDefParameter(name, type_name, self.current_direction, multi)
        self.current_reldef.params.append(rel_param)

    def finish(self, spec: "Specification"):
        """Check the relation definitions and add them to the result specification."""
        reldef_texts: Dict[
            str, "Token"
        ] = {}  # Map of defined names for relation definitions to their token.
        assert self.rel_defs is not None
        for rel_def in self.rel_defs:
            # Verify unique name.
            reldef_text = rel_def.name.tok_text
            if reldef_text in reldef_texts:
                locs = [
                    rel_def.name.get_location(),
                    reldef_texts[reldef_text].get_location(),
                ]
                self.diag_store.add(
                    diagnostics.E200(
                        reldef_text, "relation definition", location=locs[0], dupes=locs
                    )
                )
                continue

            reldef_texts[reldef_text] = rel_def.name

            # Verify parameters.
            multi_value_params: Dict[str, List[relations.RelationDefParameter]] = defaultdict(
                list
            )  # Multi-value params in each direction.
            param_texts: Dict[str, "Token"] = {}  # Map of defined parameter names to their token.
            for param in rel_def.params:
                # Register multi-value params.
                if param.multi:
                    multi_value_params[param.direction].append(param)

                # Check unique name.
                param_text = param.name.tok_text
                if param_text in param_texts:
                    locs = [
                        param.name.get_location(),
                        param_texts[param_text].get_location(),
                    ]
                    self.diag_store.add(
                        diagnostics.E200(
                            param_text,
                            "parameter definition",
                            location=locs[0],
                            dupes=locs,
                        )
                    )
                    # Continue anyway

                param_texts[param_text] = param.name

                # check type.
                type_text = param.type_name.tok_text
                typedef = spec.types.get(type_text)
                if typedef is None:
                    loc = param.type_name.get_location()
                    self.diag_store.add(diagnostics.E203("type", name=type_text, location=loc))

                    param.type = None
                else:
                    param.type = typedef.type

            # Verify lack of more than one multi-value parameter in each direction.
            found_fatal = False
            for mv_params in multi_value_params.values():
                if len(mv_params) > 1:
                    locs = [mvp.name.get_location() for mvp in mv_params]
                    direction_text = {
                        INPUT: "requiring",
                        OUTPUT: "returning",
                        INPOUT: "relating",
                    }[mv_params[0].direction]
                    self.diag_store.add(
                        diagnostics.E213(
                            f"'{direction_text}' multi-value parameter",
                            len(mv_params),
                            "at most 1",
                            location=locs[0],
                            occurrences=locs,
                        )
                    )
                    found_fatal = True

            if found_fatal:
                continue  # Fatal error.

        spec.rel_defs = self.rel_defs
        self.rel_defs = None  # Avoid adding more relation definitions.



"""Base class for goal and transformation processing to improve code sharing."""
from collections import defaultdict
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Tuple, Union

from raesl.compile import diagnostics
from raesl.compile.ast import components, exprs
from raesl.compile.typechecking.utils import resolve_var_param_node

if TYPE_CHECKING:
    from raesl.compile.ast.components import Flow, SubClause
    from raesl.compile.ast.components import ComponentInstance, VarParam
    from raesl.compile.scanner import Token


class GoalTransformBaseBuilder:
    """Common base class for checking goals and transformations.

    Arguments:
        diag_store: Storage for found diagnostics.

    Attributes:
        reported_names: Names of flows with a reported error, to avoid duplicate error
            generation.
    """

    def __init__(self, diag_store: diagnostics.DiagnosticStore):
        self.diag_store = diag_store
        self.reported_names: Set[str] = set()

    def verify_verb_prepos(
        self, verb_tok: "Token", prepos_tok: "Token", vpps: Set[Tuple[str, str]]
    ):
        """Verify verb and pre-position and report an error if the combination does not
        exist.

        Arguments:
            verb_tok: Token holding the verb text.
            prepos_tok: Token holding the prepos text.
            vpps: Available combinations of verbs and prepositions.
        """
        vpp = (verb_tok.tok_text, prepos_tok.tok_text)
        if vpp not in vpps:
            loc = verb_tok.get_location()
            self.diag_store.add(diagnostics.E211(vpp[0], vpp[1], location=loc))

    def verify_flows(self, flows: List["Flow"], vps: Dict[str, "VarParam"]) -> bool:
        """Check that each flow exists as variable or parameter. Update the link
        in the Flow object to point to the matching variable or parameter.

        Arguments:
            flows: Flows to check.
            vps: Available variables and parameters in the component.

        Returns:
            Whether all flows can be matched to a variable or parameter.
        """
        is_good = True
        for flow in flows:
            node = resolve_var_param_node(flow.name_tok, vps, self.reported_names, self.diag_store)
            if node is None:
                is_good = False
            else:
                flow.flow_node = node

        return is_good

    def resolve_component(
        self, compinst_tok: "Token", cinsts: Dict[str, "ComponentInstance"]
    ) -> Optional["ComponentInstance"]:
        """Find a component instance with the provided instance name. If it exists,
        return it, else report an error and return None, indicating failure.
        """
        compinst = cinsts.get(compinst_tok.tok_text)
        if compinst is None:
            loc = compinst_tok.get_location()
            self.diag_store.add(
                diagnostics.E203("component instance", name=compinst_tok.tok_text, location=loc)
            )

        return compinst

    def check_form(
        self,
        sect_name: str,
        kind: str,
        doesaux: "Token",
        sub_clauses: List["SubClause"],
    ):
        """Check whether the requirement or constraint form of the text is correct with
        respect to the containing section.

        Arguments:
            sect_name: Name of the section (goal or transformation).
            kind: Kind of section containing the text (requirement or constraint).
            doesaux: Token in the formulation that is either 'does' or one of the
                auxiliary verbs.
            sub_clauses: Sub clauses belong to the requirement or constraint.
        """
        # None of the possible diagnostics is considered to be fatal, so nothing is
        # returned.

        # Verify uniqueness of subclause labels.
        subclauses_by_label: Dict[str, List["SubClause"]] = defaultdict(list)
        for sub in sub_clauses:
            subclauses_by_label[sub.label_tok.tok_text].append(sub)
        for clauses in subclauses_by_label.values():
            if len(clauses) > 1:
                locs = [sub.label_tok.get_location() for sub in clauses]
                name = clauses[0].label_tok.tok_text
                self.diag_store.add(
                    diagnostics.E200(name, "subclause label", location=locs[0], dupes=locs)
                )

        # Check kind-specific requirements.
        if kind == components.CONSTRAINT:
            if doesaux.tok_type != "DOES_KW":
                loc = doesaux.get_location()
                self.diag_store.add(
                    diagnostics.E212(
                        "constraint",
                        doesaux.tok_text,
                        "'does'",
                        name=sect_name,
                        location=loc,
                    )
                )

            for sub in sub_clauses:
                self._check_constraint_expr(sub.expr)

        else:
            assert kind == components.REQUIREMENT
            if doesaux.tok_type == "DOES_KW":
                loc = doesaux.get_location()
                self.diag_store.add(
                    diagnostics.E212(
                        "requirement",
                        doesaux.tok_text,
                        "one of 'must', 'shall', 'should', 'could', or 'won't'",
                        name=sect_name,
                        location=loc,
                    )
                )

            for sub in sub_clauses:
                self._check_requirement_expr(sub.expr)

    def _check_constraint_expr(
        self,
        expr: Union[exprs.Disjunction, exprs.RelationComparison, exprs.ObjectiveComparison],
    ):
        """Check whether the subclauses have the proper form for a constraint section.
        Report an error if something wrong is found.
        """
        if isinstance(expr, exprs.Disjunction):
            for child in expr.childs:
                self._check_constraint_expr(child)

        elif isinstance(expr, exprs.RelationComparison):
            if expr.isaux_tok.tok_type != "IS_KW":
                loc = expr.isaux_tok.get_location()
                self.diag_store.add(
                    diagnostics.E212("subclause", expr.isaux_tok.tok_text, "'is'", location=loc)
                )

            if expr.math_compare == "~":
                loc = expr.cmp_tok.get_location()
                self.diag_store.add(
                    diagnostics.E212("subclause", "approximately", "hard limits", location=loc)
                )

        elif isinstance(expr, exprs.ObjectiveComparison):
            loc = expr.aux_tok.get_location()
            if expr.maximize:
                obj_text = "maximize"
            else:
                assert expr.minimize
                obj_text = "minimize"
            self.diag_store.add(
                diagnostics.E212("subclause", obj_text, "hard limits", location=loc)
            )

        else:
            assert False, "Found unexpected expression node :" + repr(expr)

    def _check_requirement_expr(
        self,
        expr: Union[exprs.Disjunction, exprs.RelationComparison, exprs.ObjectiveComparison],
    ):
        """Check whether the subclauses have the proper form for a requirement section.
        Report an error if something wrong is found.
        """
        if isinstance(expr, exprs.Disjunction):
            for child in expr.childs:
                self._check_requirement_expr(child)

        elif isinstance(expr, exprs.RelationComparison):
            if expr.isaux_tok.tok_type == "IS_KW":
                loc = expr.isaux_tok.get_location()
                self.diag_store.add(
                    diagnostics.E212(
                        "subclause",
                        "is",
                        "'must/shall/should/could/won't be'",
                        location=loc,
                    )
                )

        elif isinstance(expr, exprs.ObjectiveComparison):
            pass  # Is OK in a requirement.

        else:
            assert False, "Found unexpected expression node :" + repr(expr)



"""ESL Language Server Command Line interface."""

from typing import Optional

import click
from raesl import logger


@click.command("serve")
@click.option(
    "--port",
    "-p",
    default=None,
    type=click.INT,
    help="An optional TCP port to run on, useful for debugging.",
)
def serve(port: Optional[int] = None):
    """Start the ESL Language Server."""
    try:
        from raesl.server.server import ls
    except ImportError:
        logger.error("Missing dependencies. Make sure the 'server' extras are installed.")
        return

    if port is None:
        logger.info("Starting ESL Language Server on STDIO...")
        ls.start_io()
    else:
        logger.info(f"Starting ESL Language Server on port {port}...")
        ls.start_tcp("localhost", port)


if __name__ == "__main__":
    serve()



"""Code to construct line matching state machines.

The entry point to create a state machine is calling 'create' on a
StateMachineBuilder instance.

A state machine has locations and edges.

One location in a machine may have an 'initial' option denoting it is the first
location of the machine. Locations may also be accepting, denoting it is a
proper end state. The accepting option takes a name, making it possible to
distinguish which end location of the state machine was reached.

Edges start at a location and end at a location. Edges have a token-name, and
may only be chosen when the input has a token with the same name. An edge may
also have a tag option. The tag-name represents what kind of relevant token it
is. When an edge is taken with a tag-name, the token matching with the edge is
appended to a list associated with the tag-name. In this way, it is possible to
extract relevant information from the matched sentence afterwards.

State machine execution assumes the machine is deterministic. From a location,
each outgoing edge must have a different token-name. Note that a scanner may
not map unique text to a token. The latter is resolved by sorting the edges from
most specific token-name to least specific token-name.


In general, writing code to build locations and edges is bulky and hard to
understand. Instead a text format is defined to allow writing the state
machines as text, and let the computer build the state machine from the
description. The text format accepted by the builder is:

    state-machine ::=
      MACHINE-NAME ":" { loc-def | edge-def }+

    loc-def ::=
      LOC-NAME [ "initial" ] [ "accept" "=" ACCEPT-name ] ";"

    edge-def ::=
      LOC-NAME "->" LOC-NAME "[" TOKEN-NAME "]" [ "tag" "=" TAG-NAME ] ";"

Locations in 'loc-def' must not exist already. Missing locations in 'edge-def'
are silently created. Token names are defined in parsing/scanner.py.
"""
from typing import List, Optional, Tuple, Union

import sly
from raesl import logger
from raesl.compile.machine_files import typing
from raesl.compile.state_machine import Edge, Location, StateMachine
from sly import Lexer, Parser

# Pylint just hates Sly:
# pylint: disable=invalid-name, missing-docstring, no-self-use
# pylint: disable=function-redefined, unused-argument, used-before-assignment
# pylint: disable=undefined-variable, unsupported-assignment-operation
# flake8: noqa
# mypy: ignore-errors


class StateMachineLexer(Lexer):
    """Lexer for tokenizing state machine descriptions.

    Note that SLY and pylint are known not to like each other.
    """

    tokens = {
        NAME,
        ACCEPT_KW,
        INITIAL_KW,
        TAG_KW,
        COMMA_TK,
        SEMICOL_TK,
        EQ_TK,
        ARROW_TK,
        COLON_TK,
        SQ_OPEN,
        SQ_CLOSE,
    }

    ignore = " \t"
    ignore_comment = r"\#.*"

    @_(r"\n+")
    def ignore_newline(self, t):
        self.lineno += len(t.value)

    NAME = "[A-Za-z][A-Za-z0-9_]*"
    NAME["accept"] = ACCEPT_KW
    NAME["initial"] = INITIAL_KW
    NAME["tag"] = TAG_KW

    SQ_OPEN = "\\["
    SQ_CLOSE = "\\]"
    COMMA_TK = ","
    SEMICOL_TK = ";"
    EQ_TK = "="
    COLON_TK = ":"
    ARROW_TK = "->"


class StateMachineParser(Parser):
    tokens = StateMachineLexer.tokens

    def error(self, t):
        if t is None:
            print("Unexpected EOF encountered.")
        else:
            print("Syntax error at line {}, token {}, text={}".format(t.lineno, t.type, t.value))

    @_("NAME COLON_TK mlines")
    def machine(self, p):
        return (p[0], p[2])

    @_("")
    def mlines(self, p):
        return []

    @_("mlines loc_line")
    def mlines(self, p):
        return p[0] + [p[1]]

    @_("mlines edge_line")
    def mlines(self, p):
        return p[0] + [p[1]]

    # Locations

    @_("NAME SEMICOL_TK")
    def loc_line(self, p):
        return ("loc", p[0], [])

    @_("NAME loc_options SEMICOL_TK")
    def loc_line(self, p):
        return ("loc", p[0], p[1])

    @_("loc_option")
    def loc_options(self, p):
        return [p[0]]

    @_("loc_options COMMA_TK loc_option")
    def loc_options(self, p):
        return p[0] + [p[2]]

    @_("INITIAL_KW")
    def loc_option(self, p):
        return ("initial",)

    @_("ACCEPT_KW EQ_TK NAME")
    def loc_option(self, p):
        return ("accept", p[2])

    # Edges

    @_("NAME ARROW_TK NAME SQ_OPEN NAME SQ_CLOSE SEMICOL_TK")
    def edge_line(self, p):
        return ("edge", p[0], p[2], p[4], [])

    @_("NAME ARROW_TK NAME SQ_OPEN NAME SQ_CLOSE edge_options SEMICOL_TK")
    def edge_line(self, p):
        return ("edge", p[0], p[2], p[4], p[6])

    @_("edge_option")
    def edge_options(self, p):
        return [p[0]]

    @_("edge_options COMMA_TK edge_option")
    def edge_options(self, p):
        return p[0] + [p[2]]

    @_("TAG_KW EQ_TK NAME")
    def edge_option(self, p):
        return ("tag", p[2])


class ProcessingStateMachine(StateMachine):
    """Extended StateMachine that also holds a callback function to add extracted
    parsing information into the ast.

    Arguments:
        name: Name of the state machine, also the name of the matched sequence.
        processing_func: If not None, function that inserts relevant information from
            the matched line into the ast that is constructed.
    """

    def __init__(self, name: str, processing_func: Optional[typing.ProcessingFunc] = None):
        super().__init__(name)
        self.processing_func = processing_func


class StateMachineBuilder:
    """Class for easily constructing a state machine from a textual description.

    Attributes:
        lexer: Lexer to tokenize the input text.
        parser: Parser to interpret the tokens.
        machine: State machine to be built.
        locs: Temporary location map filled while building the state machine.
    """

    def __init__(self):
        self.lexer = StateMachineLexer()
        self.parser = StateMachineParser()
        self.machine = None
        self.locs = None

    def ensure_loc(self, name: str) -> Location:
        """Make sure a location with the provided name exists. If it doesn't, create it.

        Arguments:
            name: Name of the location to ensure.

        Returns:
            The location with the provided name.
        """
        loc = self.locs.get(name)
        if loc is not None:
            return loc

        return self.create_loc(name, [])

    def create_loc(self, name: str, opts: List[Union[Tuple[str], Tuple[str, str]]]) -> Location:
        """Create a new location.

        Arguments:
            name: Name of the location to create.
            opts: Location options, list of ('initial',) and/or ('accept', str) .

        Returns:
            The created location.
        """
        initial = False
        accept = None
        for opt in opts:
            if opt[0] == "initial":
                initial = True
            elif opt[0] == "accept":
                accept = opt[1]
            else:
                assert False, "Unexpected loc option " + repr(opt)

        loc = Location(name, accept)
        assert name not in self.locs
        self.locs[name] = loc
        if initial:
            self.machine.initial_loc = loc
        return loc

    def create(
        self, machine_text: str, processing_func: Optional[typing.ProcessingFunc] = None
    ) -> ProcessingStateMachine:
        """Create a state machine by interpreting the provided state machine text.

        Arguments:
            machine_text: Description of the state machine as defined by
                StateMachineLexer and StateMachineParser.
            processing_func: Optional processing function to copy relevant
                information into the abstract syntax tree.

        Returns:
            The state machine that implements the description.
        """
        try:
            parsed = self.parser.parse(self.lexer.tokenize(machine_text))
        except sly.lex.LexError as ex:
            logger.error(f"LEX ERROR: {ex}")
            parsed = None

        # DEBUG: Dump input text with line numbers is parsing of input text fails.
        if not parsed:
            for i, line in enumerate(machine_text.split("\n")):
                logger.debug(f"{i + 1:3d}: {line}")

        assert parsed

        mname, mlines = parsed
        self.machine = ProcessingStateMachine(mname, processing_func)
        self.locs = {}

        for mline in mlines:
            if mline[0] == "loc":
                loc_name = mline[1]
                loc_opts = mline[2]
                self.create_loc(loc_name, loc_opts)

            elif mline[0] == "edge":
                edge_source = mline[1]
                edge_dest = mline[2]
                edge_toktype = mline[3]
                edge_opts = mline[4]

                source = self.ensure_loc(edge_source)
                dest = self.ensure_loc(edge_dest)
                tag = None
                for opt in edge_opts:
                    assert opt[0] == "tag"
                    tag = opt[1]
                    break

                edge = Edge(dest, edge_toktype, tag)
                source.out_edges.append(edge)

            else:
                assert False, "Unexpected machine line " + repr(mline)

        # Some sanity checks.
        assert self.machine.initial_loc  # There should be an initial location.

        self.machine.sort_edges()
        machine = self.machine
        self.machine = None
        self.locs = None
        return machine



"""GraphView view functions for ESL."""

from copy import deepcopy
from typing import List, Optional, Tuple

from raesl.plot import utils
from raesl.plot.generic import Style
from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node


def multi_domain(
    graph: Graph,
    node_kinds: Optional[List[str]] = None,
    edge_kinds: Optional[List[str]] = None,
    edge_labels: Optional[List[str]] = None,
    edge_weights: Optional[List[str]] = None,
    lead_components: Optional[List[Node]] = None,
    depth: int = 2,
) -> Tuple[List[Node], List[Edge]]:
    """Create multi-domain-matrix visualization based on an ESL derived graph.

    Arguments:
        graph: Input graph object derived from an ESL specification.
        node_kinds: The node kinds that are included in the plot. Defaults to all
            node_kinds present within the graph.
        edge_kinds: The edge kinds that are included in the plot. Defaults to all
            edge_kinds present within the graph.
        edge_labels: The edge labels that are shown in the plot. Defaults to all
            edge_labels present within the graph.
        edge_weights: The edge weight types that are shown in the plot. Defaults to a
            single random edge weight present within the graph.
        lead_components: The lead components to be used to select other nodes.
        depth: Depth up to which lead components must be selected if no lead components
            are provided.

    Returns:
        Node list and Edge list.
    """

    if lead_components:
        utils.check_tree_disjunction(nodes=lead_components)
    else:
        lead_components = list(
            utils.get_up_to_depth(
                roots=[r for r in graph.roots if r.kind == "component"], depth=depth
            )
        )

    node_kinds = graph.node_kinds if node_kinds is None else node_kinds
    edge_kinds = graph.edge_kinds if edge_kinds is None else edge_kinds
    edge_labels = graph.edge_labels if edge_labels is None else edge_labels
    edge_weights = graph.edge_weight_labels if edge_weights is None else edge_weights

    nodes = utils.filter_nodes(graph, lead_components=lead_components, node_kinds=node_kinds)

    edges = [
        e for e in graph.edges_between_all(sources=nodes, targets=nodes) if e.kind in edge_kinds
    ]

    # Ensure that mapping dependencies appear below the diagonal.
    n2n = {kind: idx for idx, kind in enumerate(node_kinds)}
    modified_edges = []
    added_edges = []
    for e in edges:
        if e.kind != "mapping_dependency":
            continue
        if n2n[e.source.kind] > n2n[e.target.kind]:
            modified_edges.append(e)
            modified_edge = deepcopy(e)
            modified_edge.source = graph[e.source.name]
            modified_edge.target = graph[e.target.name]
            added_edges.append(modified_edge)

    for e in modified_edges:
        edges.remove(e)

    edges += added_edges

    if "variable" in node_kinds and "functional_dependency" in edge_kinds:
        edges += utils.migrate_edges_between_variables(graph, lead_components=lead_components)

    return nodes, edges


def hierarchy(graph: Graph, roots: List[Node], levels: int) -> Tuple[List[Node], List[Edge]]:
    """Filter nodes and create edges for drawing a hierarchical decomposition diagram.

    Arguments:
       graph: Source data graph.
       roots: Roots of the hierarchical diagram.
       levels: Number of levels to include in the diagram.

    Returns:
        List of selected nodes and list of created edges.
    """

    nodes, edges = [], []
    for r in roots:
        max_depth = r.depth + levels - 1
        nodes.append(r)
        crawled_nodes, crawled_edges = utils.crawl_descendants(r, max_depth)
        nodes.extend(crawled_nodes)
        edges.extend(crawled_edges)

    # Create nodes in which hierarchy information is stripped to prevent
    # nested plotting.
    stripped_nodes = []
    for node in nodes:
        copied_node = deepcopy(node)
        copied_node.parent = None
        copied_node.children = None
        stripped_nodes.append(copied_node)

    # Modify source and targets of edges to stripped nodes.
    node_dict = {n.name: n for n in stripped_nodes}
    for e in edges:
        e.source = node_dict[e.source.name]
        e.target = node_dict[e.target.name]

    return stripped_nodes, edges


def functional_dependency(graph: Graph, root: Node, levels: int) -> Tuple[List[Node], List[Node]]:
    """Filter method for drawing a nested functional dependency structure.

    Arguments
        graph: Instantiated ESL graph.
        root: Root node for which the dependency structure must be drawn.
        levels: Number of levels to include in the tree.

    Returns:
       List of nodes and edges to be displayed.

    Note:
       This method assumes that the data source graph is generated from an
       ESL specification.
    """

    components = list(utils.get_up_to_depth(roots=[root], depth=root.depth + levels))

    functions, edges = multi_domain(
        graph,
        node_kinds=["function_spec"],
        edge_kinds=["functional_dependency", "logical_dependency"],
        lead_components=components,
    )

    # Add dependencies between goals and components that cannot be traced down to
    # transformations.
    tree = sorted(
        set([a.name for c in components for a in c.ancestors]).union(
            set([c.name for c in components])
        )
    )
    edges += _add_component_goal_edges(graph, tree, functions)

    return components + functions, edges


def traceability(
    graph: Graph, root: Node, levels: int, style: Style = Style()
) -> Tuple[List[Node], List[Edge]]:
    """Filter nodes and edges for drawing a traceability diagram.

    Arguments:
        graph: Instantiated ESL graph.
        root: The root transformation specification.
        levels: Number of levels to include in the traceability diagram.
        style: RaESL style options.

    Returns:
        List of selected nodes and list of selected edges.
    """
    # Get functions in tree and traceability edges.
    g = deepcopy(graph)
    functions, edges = utils.get_function_tree(g, root, levels)

    components = [c for c in set([g[f.annotations.esl_info["body"]["active"]] for f in functions])]

    # Remove component hierarchy to make traceability hierarchy leading.
    for c in components:
        c.parent = None
        c.children = []

    if style.diagram.show_function_dependencies:
        edges.update(
            set(
                [
                    e
                    for e in g.edges_between_all(functions, functions)
                    if e.kind in ["functional_dependency", "logical_dependency"]
                ]
            )
        )

    return components + list(functions), list(edges)


def function_chain(
    graph: Graph,
    start_points: List[Tuple[Node, List[Node]]],
    end_points: List[Tuple[Node, List[Node]]],
    levels: int = 1,
    style: Style = Style(),
) -> Tuple[List[Node], List[Edge]]:
    """Filter nodes and edges for a function chain diagram.

    Arguments:
        graph: Instantiated ESL graph.
        start_points: List of start-points for the function chains.
        end_points: List of end-points for the function chains.
        levels: Number of levels to decompose intermediate components into (if present).
        style: style options class.

    Returns:
        List of selected nodes and list of selected edges.
    """
    components = set(
        utils.select_components_for_function_path(
            graph, start_points=start_points, end_points=end_points, levels=levels
        )
    )

    _, edges = multi_domain(
        graph,
        node_kinds=["function_spec"],
        edge_kinds=["functional_dependency", "logical_dependency"],
        lead_components=list(components),
    )

    sfs = []
    for p in start_points:
        sfs.extend(p[1])
    efs = []
    for p in end_points:
        efs.extend(p[1])

    path_functions = set()
    path_components = set()
    migrated_goals = []

    for path in utils.get_paths_between_all(sfs, efs, edges=edges):
        for fname in path:
            f = graph[fname]
            path_functions.add(f)
            active = graph[f.annotations.esl_info["body"]["active"]]
            if active in components:
                path_components.add(active)
            else:
                des = components.intersection(set(active.descendants))
                migrated_goals.append(f)
                for d in des:
                    if graph.edge_dict.get(d.name):
                        if graph.edge_dict[d.name].get(f.name):
                            path_components.add(d)

    sliced = graph.get_graph_slice(nodes=path_components.union(path_functions))

    if style.diagram.show_hierarchy:
        utils.rebuild_hierarchical_structure(graph, sliced)

    return sliced.leafs, [
        e
        for e in sliced.edges
        if e.source.kind == "function_spec" and e.target.kind == "function_spec"
    ]


def functional_context(
    graph: Graph, root: Node, degree: int, style: Style = Style()
) -> Tuple[List[Node], List[Edge]]:
    """Filter nodes and edges for drawing a functional context diagram.

    Arguments:
        graph: Instantiated ESL graph.
        root: The root node for which the functional context diagram must be
            drawn.
        degree: Degree up to which neighbors of neighbors must be sought.
        style: RaESL style options.

    Returns:
        List of selected nodes and list of selected edges.
    """
    neighbors = utils.get_neighbors(graph, root, root, node_kinds={"component"}, degree=degree)

    components = []
    tree = set()
    tree.add(root.name)
    if style.diagram.show_root_children and root.children:
        components.extend(root.children)
        tree.update(set([c.name for c in root.children]))
    else:
        components.append(root)

    tree.update(set([n.name for n in neighbors]))
    if style.diagram.show_neighbor_children:
        for neighbor in neighbors:
            if neighbor.children:
                components.extend(neighbor.children)
                tree.update(set([c.name for c in neighbor.children]))
            else:
                components.append(neighbor)
    else:
        components.extend(neighbors)

    functions, edges = multi_domain(
        graph,
        node_kinds=["function_spec"],
        edge_kinds=["functional_dependency", "logical_dependency"],
        lead_components=components,
    )

    edges += _add_component_goal_edges(graph, sorted(tree), functions)

    return components + functions, edges


def _add_component_goal_edges(graph: Graph, tree: List[Node], functions: List[Node]) -> List[Edge]:
    """Create edges between component or cluster nodes if a goal cannot be related
    to a transformation:

    Arguments:
        graph: Instantiated ESL graph.
        tree: List of components that are within the (sub)tree that is plotted.
        functions: List of function specs that is plotted.

    Returns:
        List of edges.
    """
    edges = []
    goals = [f for f in functions if f.annotations.esl_info["sub_kind"] == "goal"]
    transformations = [
        f for f in functions if f.annotations.esl_info["sub_kind"] == "transformation"
    ]
    for goal in goals:
        if (
            not list(graph.edges_between_all(sources=[goal], targets=transformations))
            and goal.annotations.esl_info["body"]["passive"] in tree
        ):
            # No destination transformation is found. Add edge to destination component
            # if the destination component is part of the selected components.
            for e in graph.directed_edges[goal.annotations.esl_info["body"]["passive"]][goal.name]:
                edges.append(
                    Edge(
                        source=e.target,
                        target=e.source,
                        kind=e.kind,
                        labels=e.labels,
                        weights=e.weights,
                        annotations=e.annotations,
                    )
                )

        if (
            not list(graph.edges_between_all(sources=transformations, targets=[goal]))
            and goal.annotations.esl_info["body"]["active"] in tree
        ):
            # No source transformation is found. Add edge from to source component.
            edges.extend(
                graph.directed_edges[goal.annotations.esl_info["body"]["active"]][goal.name]
            )

    return edges



"""Language Server class module."""
from collections import defaultdict
from pathlib import Path
from typing import List, Optional

from pygls import lsp, server
from raesl import logger
from raesl.server.config import EslConfig
from raesl.utils import cleanup_path, path_to_uri

# Compile module is excluded during docs generation.
try:
    from raesl.compile import diagnostics, parser, scanner
    from raesl.compile.instantiating import graph_building
except ImportError:
    pass


class EslServer(server.LanguageServer):
    """ESL Language Server."""

    CONFIG = "esl"  # Config section in editors.
    CMD_COMPILE_PATH = "esl.compilePath"

    def __init__(self):
        super().__init__()
        self.config = EslConfig()

    async def update_config(self):
        """Handle a config update request."""
        self.show_message_log("Refreshing server config...")
        try:
            cfg = await self.get_configuration_async(
                lsp.ConfigurationParams([lsp.ConfigurationItem("", self.CONFIG)])
            )
            self.config.parse_config(self.workspace, cfg[0])
            self.show_message_log("Refreshed server config.")
        except Exception as e:
            self.show_message_log(f"Error occurred while updating config: '{e}'.")


ls = EslServer()


@ls.feature(lsp.TEXT_DOCUMENT_DID_OPEN)
async def did_open(ls: EslServer, params: lsp.DidOpenTextDocumentParams):
    """Opened document handling."""
    ls.show_message_log(f"Opened: '{params.textDocument.uri}'")
    await ls.update_config()
    _validate(ls, ls.config.get_paths(ls.workspace, params.textDocument.uri))


@ls.feature(lsp.TEXT_DOCUMENT_DID_CLOSE)
async def did_close(ls: EslServer, params: lsp.DidCloseTextDocumentParams):
    """Closed document handling."""
    ls.show_message_log(f"Closed: '{params.textDocument.uri}'")
    await ls.update_config()
    paths = ls.config.get_paths(ls.workspace) or []
    if cleanup_path(params.textDocument.uri) not in paths:
        ls.publish_diagnostics(doc_uri=params.textDocument.uri, diagnostics=[])
    _validate(ls, paths)


@ls.feature(lsp.TEXT_DOCUMENT_DID_SAVE)
async def did_save(ls: EslServer, params: lsp.DidSaveTextDocumentParams):
    """Saved document handling."""
    ls.show_message_log(f"Saved: '{params.textDocument.uri}'")
    await ls.update_config()
    _validate(ls, ls.config.get_paths(ls.workspace, params.textDocument.uri))


@ls.feature(lsp.WORKSPACE_DID_CHANGE_WATCHED_FILES)
async def did_change_watched(ls: EslServer, params: lsp.DidChangeWatchedFilesParams):
    """Changed watched files handling."""
    ls.show_message_log(f"Changed: '{params.changes}")
    await ls.update_config()
    _validate(ls, ls.config.get_paths(ls.workspace))


@ls.feature(lsp.WORKSPACE_DID_CHANGE_CONFIGURATION)
async def did_change_config(ls: EslServer, params: lsp.DidChangeConfigurationParams):
    """Config changed handling."""
    ls.show_message_log("Workspace config changed.")
    await ls.update_config()
    _validate(ls, ls.config.get_paths(ls.workspace))


@ls.feature("$/setTraceNotification")
async def set_trace_notification(ls: EslServer, *args):
    """Unkown notification. Probably a config change."""
    ls.show_message_log("Unknown notification.")
    await ls.update_config()
    _validate(ls, ls.config.get_paths(ls.workspace))


def _validate(ls: EslServer, paths: Optional[List[Path]] = None) -> bool:
    """Validate ESL files on given paths and push diagnostics."""
    if not paths:
        logger.info("No paths to validate.")
        return True
    try:
        logger.info(f"Validating '{paths}'...")
        # Create specification object.
        diag_store, spec = parser.parse_spec(
            scanner.Lexer(str(p), p.read_text(), 0, 0, 0, []) for p in paths
        )
        if diag_store.has_severe() or spec is None:
            _publish(ls, paths, diag_store)
            logger.error("Parsing failed validation.")
            return False

        # Derive dependency graph.
        graph = graph_building.GraphFactory(diag_store=diag_store, spec=spec).make_graph()
        if diag_store.has_severe() or graph is None:
            _publish(ls, paths, diag_store)
            logger.error("Graph building failed validation.")
            return False
        elif ls.config.output is not None:
            ls.show_message_log(f"Dumping output graph to '{ls.config.output}'...")
            from ragraph.io.json import to_json

            to_json(graph, path=ls.config.output)
            ls.show_message_log("Dumped output graph.")

        # Publish any non-severe diagnostics.
        _publish(ls, paths, diag_store)
        logger.info("Successful validation.")
        return True

    except Exception as e:
        ls.show_message(
            f"Unexpected error:\n{e}",
            msg_type=lsp.MessageType.Error,
        )
        return False


def _publish(ls: EslServer, paths: List[Path], diag_store: diagnostics.DiagnosticStore):
    """Publish diagnostics from the Diagnostic Store via the language server."""
    ls.show_message_log(f"Publishing {len(diag_store.diagnostics)} diagnostic(s)...")

    # Fetch diagnostics per URI and cast to pygls Diagnostic.
    diag_map = defaultdict(list)
    for diag in diag_store.diagnostics:
        if diag.related_information:
            related = [
                lsp.DiagnosticRelatedInformation(info.location, info.message)
                for info in diag.related_information
            ]
        else:
            related = None

        diagnostic = lsp.Diagnostic(
            range=lsp.Range(diag.range.start, diag.range.end),
            message=diag.message,
            severity=lsp.DiagnosticSeverity[diag.severity.name],
            code=diag.code,
            source=diag.source,
            related_information=related,
        )

        uri = path_to_uri(diag.location.uri)
        diag_map[uri].append(diagnostic)

    # Publish diagnostics per URI.
    for path in paths:
        p = path_to_uri(path)
        ls.publish_diagnostics(doc_uri=p, diagnostics=diag_map[p])

    ls.show_message_log("Published diagnostic(s).")



"""State machine files for recognizing lines and ESL."""
from typing import Dict, List, Optional, Tuple

from raesl.compile.machine_files import (
    behavior,
    builder,
    comments,
    component_definitions,
    component_instances,
    designs,
    goals,
    groups,
    needs,
    parameters,
    relation_definitions,
    relation_instances,
    transforms,
    type_defs,
    typing,
    variables,
    verb_defs,
)

# If not None, cached dict of line machine name to the state machine object.
_LINE_MACHINES: Optional[Dict[str, builder.ProcessingStateMachine]] = None


def collect_line_machines() -> Dict[str, builder.ProcessingStateMachine]:
    """Construct and collect all available line machine, and return them."""
    global _LINE_MACHINES

    if _LINE_MACHINES is not None:
        return _LINE_MACHINES

    _LINE_MACHINES = {}
    mach_builder = builder.StateMachineBuilder()

    collections: List[List[Tuple[str, str, Optional[typing.ProcessingFunc]]]] = [
        behavior.MACHINES,
        comments.MACHINES,
        component_definitions.MACHINES,
        component_instances.MACHINES,
        designs.MACHINES,
        goals.MACHINES,
        groups.MACHINES,
        needs.MACHINES,
        parameters.MACHINES,
        relation_definitions.MACHINES,
        relation_instances.MACHINES,
        transforms.MACHINES,
        type_defs.MACHINES,
        variables.MACHINES,
        verb_defs.MACHINES,
    ]

    for mach_collection in collections:
        for name, spec_text, proc_func in mach_collection:
            mach = mach_builder.create(spec_text, proc_func)
            _LINE_MACHINES[name] = mach

    return _LINE_MACHINES



"""Expressions to store and reason about values and boundaries."""
from typing import TYPE_CHECKING, Optional, Sequence, Set

if TYPE_CHECKING:
    from raesl.compile.ast.nodes import VarNode
    from raesl.compile.scanner import Token
    from raesl.types import Location


class DataValue:
    """Some kind of data value. Do not use this, but use a derived class instead."""

    def get_units(self) -> Optional[Set[str]]:
        """Obtain the units of the value. Gives a set of names without square brackets
        where lack of units results in the empty set, and not supporting units gives
        None.

        Returns:
            Names of the available units without square brackets or None.
        """
        raise NotImplementedError("Implement me in {}".format(repr(self)))


class Value(DataValue):
    """A value with an optional unit. Don't modify these objects in-place, create a new
    object instead.

    Arguments:
        value: Stored value as text.
        unit: Either None or text describing the unit. Treat as read-only, as changing
            it may break the cache.

    Attributes:
        _unit_cache: Units of the literal after normalizing self.unit. Computed on
            demand.
    """

    def __init__(self, value: "Token", unit: Optional["Token"] = None):
        super(Value, self).__init__()
        self.value = value
        self.unit = unit
        self._unit_cache: Optional[Set[str]] = None  # Lazily computed.

        assert self.unit is None or self.unit.tok_text not in ("", "[]")

    def get_units(self) -> Optional[Set[str]]:
        # Updates self._unit_cache on first call.
        if self._unit_cache is None:
            if self.unit is None:
                self._unit_cache = set(["-"])
            elif self.unit.tok_text.startswith("[") and self.unit.tok_text.endswith("]"):
                self._unit_cache = set([self.unit.tok_text[1:-1]])
            else:
                self._unit_cache = set([self.unit.tok_text])

        return self._unit_cache

    def __eq__(self, other):
        if not isinstance(other, Value):
            return False
        if self.value.tok_text != other.value.tok_text:
            return False
        if self.unit is None and other.unit is None:
            return True
        if self.unit is not None and other.unit is not None:
            return self.unit.tok_text == other.unit.tok_text
        return False

    def __repr__(self):
        if self.unit is None:
            return "Value({})".format(self.value.tok_text)
        return "Value({}, {})".format(self.value.tok_text, self.unit.tok_text)


class VariableValue(DataValue):
    """Class representing a variable or parameter as value.

    Arguments:
        var_tok: Token stating the possibly dotted name of the variable.
        var_node: If not None, the node represented by the object. Set during type
            checking.
    """

    def __init__(self, var_tok: "Token"):
        self.var_tok = var_tok

        self.var_node: Optional["VarNode"] = None

    def get_units(self) -> Optional[Set[str]]:
        assert self.var_node is not None
        return self.var_node.the_type.get_units()


class Expression:
    """Base class of an expression."""


class Comparison(Expression):
    """Class storing a comparison.

    Arguments:
        is_constraint: Whether the comparison is considered to be a constraint rather
            than a requirement.
    """

    def __init__(self, is_constraint: bool):
        super(Comparison, self).__init__()
        self.is_constraint = is_constraint


_MATH_OP_TRANSLATE = {
    "LEAST_KW": ">=",
    "MOST_KW": "<=",
    "EQUAL_KW": "==",
    "NOT_KW": "!=",
    "SMALLER_KW": "<",
    "GREATER_KW": ">",
    "APPROXIMATELY_KW": "~",
}


class RelationComparison(Comparison):
    """A relation between a variable and either a value or a variable.

    Arguments:
        is_constraint: Whether the comparison is considered to be a constraint rather
            than a requirement.
        lhs_var: Left hand side variable being compared.
        isaux_tok: 'is' for a constraint, else the aux word for expressing strength
           of the comparison.
        cmp_tok: One of the key words that expression the comparison to perform.
        math_compare: Translated 'cmp_tok', with the ascii math text.
        rhs_varval: Right hand side variable or value.
    """

    def __init__(
        self,
        is_constraint: bool,
        lhs_var: VariableValue,
        isaux_tok: "Token",
        cmp_tok: "Token",
        rhs_varval: DataValue,
    ):
        super(RelationComparison, self).__init__(is_constraint)
        self.lhs_var = lhs_var
        self.isaux_tok = isaux_tok
        self.cmp_tok = cmp_tok
        self.math_compare = _MATH_OP_TRANSLATE[cmp_tok.tok_type]
        self.rhs_varval = rhs_varval

    def get_location(self) -> "Location":
        """Return a location to point at the comparison for error reporting purposes."""
        return self.isaux_tok.get_location()


class ObjectiveComparison(Comparison):
    """An intended direction for a variable. Note that the 'maximize' parameter
    controls both the 'maximize' and 'minimize' desires.

    Arguments:
        lhs_var: Variable with the objective.
        aux_tok: One of the auxiliary verbs expressing strength of the objective.
        maximize: If set the comparison expresses the desire to maximize the variable.

    Attributes:
        minimize: Opposite of maximize.
    """

    def __init__(self, lhs_var: VariableValue, aux_tok: "Token", maximize: bool):
        super(ObjectiveComparison, self).__init__(False)
        self.lhs_var = lhs_var
        self.aux_tok = aux_tok
        self.maximize = maximize

    def get_location(self) -> "Location":
        """Return a location to point at the comparison for error reporting purposes."""
        return self.aux_tok.get_location()

    @property
    def minimize(self) -> bool:
        return not self.maximize


class Disjunction(Expression):
    """Disjunctive expression (also known as 'or' expression).
    It is true iff at least one of it child expressions is true.

    Arguments:
        childs: Child expressions of the disjunction. It is recommended to have at
            least two children in an object.
    """

    def __init__(self, childs: Sequence[Expression]):
        self.childs = childs



"""Datasets module. Contains various ESL datasets (see :obj:`enum()`, :obj:`get()`,
and :obj:`info`).
"""

import importlib
import shutil
from pathlib import Path
from typing import List

HERE = Path(__file__).parent


def enum() -> List[str]:
    """Enumerate all available datasets."""
    return sorted(
        [
            d.name
            for d in HERE.iterdir()
            if d.is_dir() and not d.name.startswith(".") and not d.name.startswith("_")
        ]
    )


def check(name: str):
    """Check whether a dataset exists."""
    available = enum()
    if name not in available:
        raise ValueError(
            "Dataset {} cannot be found. Please pick one of {}.".format(name, available)
        )


def info(name: str) -> str:
    """Get information about a dataset."""
    check(name)
    mod = importlib.import_module(f"raesl.datasets.{name}")
    doc = mod.__doc__
    return str(doc)


def copy(name: str, output_path: str, as_file: bool = False) -> None:
    """Copy a dataset into a directory or file."""
    check(name)
    p = Path(output_path)
    p.parent.mkdir(exist_ok=True, parents=True)

    loc = HERE / name
    files = [f for f in (HERE / name).glob("**/*.esl")]

    if as_file:
        p.write_text("\n\n".join(f.read_text() for f in files))
    else:
        if p.exists() and any(p.iterdir()):
            raise ValueError(f"Directory {p} already exists and is not empty.")

        for f in files:
            dest = p / f.relative_to(loc)
            dest.parent.mkdir(exist_ok=True, parents=True)
            shutil.copy(f, dest)



"""ESL Pygments Lexer module.

A Pygments Lexer for the Elephant Specification Language (ESL). Mainly used for syntax
highlighting in generated documentation using Sphinx.
"""
from pygments.lexer import RegexLexer, include, words
from pygments.token import Comment, Keyword, Name, Text, Whitespace

H = r"([ \t]+)"
CON = r"(?=([ \t]*\.\.\.))"
EOT = r"(?=((\.)|(\s)|(#)))"
SOT = r"((^)|(?<=((\.)|(\s))))"
EOL = r"(?!([ \t]*\.\.\.))(?=([ \t]*($|#)))"
BOL = r"([ \t]*)"

keywords = [
    "world",
    "empty",
    "define",
    "is",
    "be",
    "a",
    "an",
    "with",
    "argument",
    "arguments",
    "goal-requirement",
    "goal-requirements",
    "transformation-requirement",
    "transformation-requirements",
    "design-requirement",
    "design-requirements",
    "behavior-requirement",
    "behavior-requirements",
    "goal-constraint",
    "goal-constraints",
    "transformation-constraint",
    "transformation-constraints",
    "design-constraint",
    "design-constraints",
    "behavior-constraint",
    "behavior-constraints",
    "need",
    "needs",
    "subclause",
    "subclauses",
    "verb",
    "verbs",
    "type",
    "types",
    "component",
    "components",
    "relation",
    "relations",
    "variable",
    "variables",
    "parameter",
    "parameters",
    "comment",
    "comments",
    "shall",
    "must",
    "could",
    "should",
    "won't",
    "or",
    "and",
    "not",
    "equal to",
    "greater than",
    "smaller than",
    "at least",
    "at most",
    "approximately",
    "maximized",
    "minimized",
    ",",
    "end",
    "case",
    "when",
    "then",
    "no other case applies",
    "does",
    "unit",
    "units" "requiring",
    "relating",
    "returning",
    "one or more",
    "t.b.d.",
    "bundle of",
    "group of",
    "variable-group",
    "variable-groups",
    "enumeration of",
    "property",
]


class EslLexer(RegexLexer):
    """Elephant Specification Language Lexer."""

    name = "Elephant Specification Language"
    aliases = ["esl", "elephant"]
    filenames = [".esl"]
    tokens = {
        "comments": [(r"#.*\n", Comment)],
        "root": [
            include("comments"),
            (r"\s+", Whitespace),
            (words(keywords, prefix=SOT, suffix=EOT), Keyword.Reserved),
            (r"\.\.\.", Keyword.Reserved),
            (r"[\w-]+", Name),  # Skip current word.
            (r".", Text),  # Skip any character if everything else fails.
        ],
    }
